"""This module contains the pattern distribution as used by the generator
and generator function of the synthetic data framework."""

import json
import os
from pathlib import Path
from random import choices, randint

from pydexpi.syndata.pattern import Pattern


class PatternDistribution:
    """A PatternDistribution is a collection of Patterns that share the same
    properties with regards to the generation logic. The patterns are grouped
    for the sake of randomness and for structure. The grouping is based on
    similarity mainly regarding:
        - Outside connectors.
        - Transition logic.
        - Conceptual similarity (e.g. distillation equipment).

    Note: Pattern labels in a distribution are exclusive

    Attributes
    ----------
    name : str
        The name of the pattern distribution, preferably an umbrella term for
        the patterns contained.
    patterns : dict[str, Pattern]
        A dictionary containing the distributions patterns as values, with the
        pattern labels as keys.
    probabilities : dict[str, float]
        A dictionary containing the distributions probabilities as values, with
        the pattern labels as keys.
    connector_labels : list[str]
        The labels of the connectors that each of the patterns in the
        distribution must contain.

    """

    def __init__(
        self,
        name: str,
        patterns: dict[str, Pattern],
        probabilities: dict[str, float],
        connector_labels: set[str],
    ) -> None:
        """Constructor of a pattern distribution. Name, distribution, and
        connector_labels stored as attributes. Each pattern in the distribution
        is validated against the connector labels.

        Parameters
        ----------
        name : str
            The name of the distribution.
        patterns : dict[str, Pattern]
            A dictionary containing the distributions patterns as values, with
            the pattern labels as keys.
        probabilities : dict[str, float]
            A dictionary containing the distributions probabilities as values,
            with the pattern labels as keys.
        connector_labels : set[str]
            The labels of the connectors that each pattern needs to have to be
            part of the distribution.

        Raises
        ------
        ValueError
            In case a pattern of the distribution does not fulfil the connector
            requirements.
        """
        self.connector_labels = set(connector_labels)
        self.name = name
        if set(patterns.keys()) != set(probabilities.keys()):
            msg = "Dictionaries for patterns and probabilities must have the same keys."
            raise ValueError(msg)
        for the_label in patterns.keys():
            if patterns[the_label].label != the_label:
                msg = "Pattern labels dont match label keys"
                raise ValueError(msg)
            if not self.check_pattern_compatibility(patterns[the_label]):
                raise ValueError("Not all patterns in passed distribution compatible")
            if probabilities[the_label] < 0:
                raise ValueError("Probabilities may not be negative.")
        self.patterns = patterns
        self.probabilities = probabilities

    def add_pattern(self, pattern: Pattern, probability: float, normalize: bool = True) -> None:
        """
        Add a pattern to the distribution with a specified probability.

        Parameters
        ----------
        name: str
            Name of the new patterndistribution
        pattern : Pattern
            The pattern to be added to the distribution.
        probability : float
            The probability associated with the pattern. Must be between 0 and 1.
        normalize: bool
            Wherther the probabilities should be normalized after adding the pattern.

        Raises
        ------
        ValueError
            If the pattern does not contain all required connectors.
            If the pattern is already part of the distribution.
            If the probability is not between 0 and 1.
        """
        if not self.check_pattern_compatibility(pattern):
            msg = f"Pattern {pattern} does not contain all required connectors to be added to distribution."
            raise ValueError(msg)
        if pattern.label in self.labels():
            raise ValueError(f"Pattern with label {pattern.label} already part of distribution.")

        self.patterns[pattern.label] = pattern
        self.probabilities[pattern.label] = probability
        if normalize:
            self.normalize_probabilities()

    def normalize_probabilities(self) -> None:
        """
        Normalize the probabilities in the distribution so they sum to 1.

        This method modifies the probabilities in the distribution such that
        the total sum of all probabilities equals 1. If the total probability
        is greater than 0, each probability is divided by the total probability.

        """
        total_prob = sum(self.probabilities.values())
        self.probabilities = {
            label: prob / total_prob for label, prob in self.probabilities.items()
        }

    def check_pattern_compatibility(self, pattern: Pattern) -> bool:
        """Checks if the pattern is compatible with the pattern distribution, ie
        contains all required connector labels. Note that a pattern may have more
        connectors than the required ones to be valid.

        Parameters
        ----------
        pattern : Pattern
            The pattern to be evaluated for compatibility.

        Returns
        -------
        bool
            True if compatible, else False
        """
        return self.connector_labels.issubset(pattern.connectors.keys())

    def __iter__(self):
        """Defines iterable as iterating over the contents of the elements of
        the distribution.
        """
        for key in self.labels():
            yield self.patterns[key].copy_pattern(), self.probabilities[key]

    def sample_pattern(self) -> tuple[Pattern, float]:
        """Samples a pattern from the pattern distribution with respect to the
        probabilities of the distribution.

        Returns
        -------
        Pattern
            The sampled pattern.
        """
        pattern_choices = self.labels()
        probabilities = list(self.probabilities.values())
        sampled_key = choices(pattern_choices, weights=probabilities, k=1)[0]
        return self.patterns[sampled_key].copy_pattern(), self.probabilities[sampled_key]

    def random_pattern(self) -> tuple[Pattern, float]:
        """Samples a pattern from the distribution completely at random,
        disregarding the probabilities.

        Returns
        -------
        Pattern
            The randomly sampled pattern.
        """
        random_index = randint(0, len(self.patterns) - 1)
        random_key = list(self.labels())[random_index]
        return self.patterns[random_key].copy_pattern(), self.probabilities[random_key]

    @classmethod
    def load(cls, dir_path: Path, name: str):
        """Loads a distribution in the directory specified in path_dir with the
        name given in name, as previously saved by the save_distribution method.

        Parameters
        ----------
        path_dir : Path
            The directory the distribution is saved in.
        name : str
            The name of the distribution.

        Returns
        -------
        PatternDistribution
            The loaded pattern distribution.
        """
        distribution_directory = Path(dir_path) / name
        patterns = []
        for filename in os.listdir(distribution_directory):
            if filename == f"{name}.json":
                with open(distribution_directory / filename) as json_file:
                    load_dict = json.load(json_file)
                    connector_labels = set(load_dict["Connector labels"])
                    new_probabilities_dict = load_dict["Probabilities"]
            else:
                pattern_name = Path(filename).stem
                new_pattern = Pattern.load(distribution_directory, pattern_name)
                patterns.append(new_pattern)
        new_pattern_dict = {pattern.label: pattern for pattern in patterns}

        return cls(name, new_pattern_dict, new_probabilities_dict, connector_labels)

    def save(self, dir_path: Path):
        """Saves a distribution to a directory in dirpath with the folder name
        taken from its name attribute. All patterns are saved in a new folder
        with the distribution name. Metadata, specifically the distribution
        probabilities and connector labels, are saved along in a separate json
        in that folder.

        Parameters
        ----------
        path_dir : str
            The path of the directory the pattern is stored in
        """
        # Make a directory for the distribution
        if not os.path.exists(dir_path):
            raise FileNotFoundError("No directory found in path")
        distribution_directory = Path(dir_path) / self.name
        if not os.path.exists(distribution_directory):
            os.mkdir(distribution_directory)
        else:
            raise FileExistsError(
                "Already a file for a pattern distribution found in the path location."
            )
        save_dict = {
            "Connector labels": list(self.connector_labels),
            "Probabilities": self.probabilities,
        }
        with open(distribution_directory / f"{self.name}.json", "w") as json_file:
            json.dump(save_dict, json_file, indent=4)

        for pattern in self.patterns.values():
            pattern.save(distribution_directory, pattern.label)

    def labels(self):
        """Get the labels of the patterns in the distribution."""
        return list(self.patterns.keys())



from pydexpi.dexpi_classes import instrumentation


def add_signal_generating_function_to_instrumentation_function(
    instrumentation_function: instrumentation.ProcessInstrumentationFunction,
    signal_generating_function: instrumentation.ProcessSignalGeneratingFunction,
    measuring_line: instrumentation.MeasuringLineFunction,
) -> None:
    """
    Add a signal generating function to a process instrumentation function via a measuring line.

    Parameters
    ----------
    instrumentation_function : instrumentation.ProcessInstrumentationFunction
        The process instrumentation function that the generation function should be added to.
    signal_generating_function : instrumentation.ProcessSignalGeneratingFunction
        The function that generates the signal.
    measuring_line : instrumentation.MeasuringLineFunction
        The line that conveys the signal from the generating function to the instrumentation function.

    Raises
    ------
    ValueError
        If the measuring line already exists in the instrumentation function.
        If the signal generating function already exists in the instrumentation function.
        If the measuring line already has a source that is not the signal generating function.
        If the measuring line already has a target that is not the instrumentation function.
    """

    # Some consistency checks
    if measuring_line in instrumentation_function.signalConveyingFunctions:
        raise ValueError("The measuring line already exists in the instrumentation function.")
    if signal_generating_function in instrumentation_function.processSignalGeneratingFunctions:
        raise ValueError(
            "The signal generating function already exists in the instrumentation function."
        )
    if (
        measuring_line.source is not signal_generating_function
        and measuring_line.source is not None
    ):
        raise ValueError("The measuring line already has a source.")
    if measuring_line.target is not instrumentation_function and measuring_line.target is not None:
        raise ValueError("The measuring line already has a target.")

    # Connect measuring line
    measuring_line.source = signal_generating_function
    measuring_line.target = instrumentation_function

    # Add line and generating function to instrumentation function
    instrumentation_function.signalConveyingFunctions.append(measuring_line)
    instrumentation_function.processSignalGeneratingFunctions.append(signal_generating_function)


def add_signal_opc_to_instrumentation_function(
    instrumentation_function: instrumentation.ProcessInstrumentationFunction,
    signal_opc: instrumentation.SignalOffPageConnector,
    signal_line: instrumentation.SignalLineFunction,
) -> None:
    """
    Add a signal off-page connector to an instrumentation function via a signal line.

    The direction of the signal line is determined by the type of the signal OPC.
    If the signal OPC is a FlowOutSignalOffPageConnector, the signal line is outgoing.

    Parameters
    ----------
    instrumentation_function : instrumentation.ProcessInstrumentationFunction
        The instrumentation function to which the off-page connector is added to.
    signal_opc : instrumentation.SignalOffPageConnector
        The signal off-page connector to add.
    signal_line : instrumentation.SignalLineFunction
        The signal line that connects the instrumentation function and the off-page connector.

    Raises
    ------
    ValueError
        If the signal OPC already exists in the instrumentation function.
        If the signal line already exists in the instrumentation function.
        If the signal line already has a target (for outgoing signals).
        If the signal line already has a source (for incoming signals).
    """

    # Some consistency checks
    if signal_opc in instrumentation_function.signalConnectors:
        raise ValueError("The signal OPC already exists in the instrumentation function.")
    if signal_line in instrumentation_function.signalConveyingFunctions:
        raise ValueError("The signal line already exists in the instrumentation function.")

    is_outgoing = isinstance(signal_opc, instrumentation.FlowOutSignalOffPageConnector)
    is_incoming = isinstance(signal_opc, instrumentation.FlowInSignalOffPageConnector)
    if not is_outgoing and not is_incoming:
        raise ValueError("The signal_opc is not a valid OPC.")

    if is_outgoing:
        # Some consistency checks
        if signal_line.target is not signal_opc and signal_line.target is not None:
            raise ValueError("The signal line already has a target.")
        if signal_line.source is not instrumentation_function and signal_line.source is not None:
            raise ValueError("The signal line already has a source.")

        # Connect signal line
        signal_line.target = signal_opc
        signal_line.source = instrumentation_function

        # Add line and generating function to instrumentation function
        instrumentation_function.signalConveyingFunctions.append(signal_line)
        instrumentation_function.signalConnectors.append(signal_opc)

    else:
        if signal_line.source is not signal_opc and signal_line.source is not None:
            raise ValueError("The signal line already has a source.")
        if signal_line.target is not instrumentation_function and signal_line.target is not None:
            raise ValueError("The signal line already has a target.")

        # Connect signal line
        signal_line.source = signal_opc
        signal_line.target = instrumentation_function

        # Add line and generating function to instrumentation function
        instrumentation_function.signalConveyingFunctions.append(signal_line)
        instrumentation_function.signalConnectors.append(signal_opc)


def add_actuating_function_to_instrumentation_function(
    instrumentation_function: instrumentation.ProcessInstrumentationFunction,
    actuating_function: instrumentation.ActuatingFunction
    | instrumentation.ActuatingElectricalFunction,
    signal_line: instrumentation.SignalLineFunction,
):
    """
    Add an actuating function to an instrumentation function via a signal line.

    The actuating function can be either a regular actuating function or an electrical actuating
    function.

    Parameters
    ----------
    instrumentation_function : instrumentation.ProcessInstrumentationFunction
        The instrumentation function the actuating function is added to.
    actuating_function : instrumentation.ActuatingFunction or instrumentation.ActuatingElectricalFunction
        The actuating function that will receive the signal.
    signal_line : instrumentation.SignalLineFunction
        The signal line that conveys the signal from the instrumentation function to the actuating function.

    Raises
    ------
    ValueError
        If the signal line already exists in the instrumentation function.
        If the actuating function already exists in the instrumentation function.
        If the signal line already has a source that is not the instrumentation function.
        If the signal line already has a target that is not the actuating function.
    """
    is_electrical = isinstance(actuating_function, instrumentation.ActuatingElectricalFunction)

    # Some consistency checks
    if signal_line in instrumentation_function.signalConveyingFunctions:
        raise ValueError("The signal line already exists in the instrumentation function.")
    if is_electrical:
        if actuating_function in instrumentation_function.actuatingElectricalFunctions:
            raise ValueError(
                "The electrical actuating function already exists in the instrumentation function."
            )
    else:
        if actuating_function in instrumentation_function.actuatingFunctions:
            raise ValueError(
                "The actuating function already exists in the instrumentation function."
            )

    if signal_line.source is not instrumentation_function and signal_line.source is not None:
        raise ValueError("The signal line already has a source.")
    if signal_line.target is not actuating_function and signal_line.target is not None:
        raise ValueError("The signal line already has a target.")

    # Connect signal line
    signal_line.source = instrumentation_function
    signal_line.target = actuating_function

    # Add line and generating function to instrumentation function
    instrumentation_function.signalConveyingFunctions.append(signal_line)
    if is_electrical:
        instrumentation_function.actuatingElectricalFunctions.append(actuating_function)
    else:
        instrumentation_function.actuatingFunctions.append(actuating_function)


def connect_instrumentation_functions(
    source_function: instrumentation.ProcessInstrumentationFunction,
    target_function: instrumentation.ProcessInstrumentationFunction,
    signal_line: instrumentation.SignalLineFunction,
    source_signal_line: bool = True,
):
    """
    Connect two instrumentation functions via a signal line.

    Parameters
    ----------
    source_function : instrumentation.ProcessInstrumentationFunction
        The source instrumentation function that will send the signal.
    target_function : instrumentation.ProcessInstrumentationFunction
        The target instrumentation function that will receive the signal.
    signal_line : instrumentation.SignalLineFunction
        The signal line that conveys the signal from the source to the target function.
    source_signal_line : bool, optional
        If True, the signal line is added to the source function's signalConveyingFunctions.
        If False, the signal line is added to the target function's signalConveyingFunctions.
        Default is True.

    Raises
    ------
    ValueError
        If the signal line already exists in either the source or target function.
        If the signal line already has a source that is not the source function.
        If the signal line already has a target that is not the target function.
    """

    # Some consistency checks
    if (
        signal_line in source_function.signalConveyingFunctions
        or signal_line in target_function.signalConveyingFunctions
    ):
        raise ValueError("The signal line already exists in the instrumentation function.")
    if signal_line.source is not source_function and signal_line.source is not None:
        raise ValueError("The signal line already has a source.")
    if signal_line.target is not target_function and signal_line.target is not None:
        raise ValueError("The signal line already has a target.")

    # Connect signal line
    signal_line.source = source_function
    signal_line.target = target_function

    # Add line and generating function to instrumentation function
    if source_signal_line:
        source_function.signalConveyingFunctions.append(signal_line)
    else:
        target_function.signalConveyingFunctions.append(signal_line)



"""
Dexpi Model Toolkit
--------------------
This module contains functions and tools that concern overall of an overall
dexpi model/conceptual model.

"""

from collections.abc import Callable
from typing import Any, get_origin, get_type_hints

from pydexpi.dexpi_classes.customization import CustomAttributeOwner
from pydexpi.dexpi_classes.dexpiBaseModels import DexpiBaseModel
from pydexpi.dexpi_classes.dexpiModel import ConceptualModel, DexpiModel


def combine_dexpi_models(models: list[DexpiModel], **kwargs) -> DexpiModel:
    """This function combines the contents of a list of dexpi models and returns
    a new model with its contents.

    The combination is performed on all list attributes of the conceptual model.
    The meta data attribute and the data attributes of the dexpi model can be
    passed in via kwargs.

    Parameters
    ----------
    models : list[DexpiModel]
        The list of dexpi models to be combined.

    Returns
    -------
    DexpiModel
        The combined dexpi model.

    Raises
    ------
    NotImplementedError
        Combining diagram information is not yet implemented. If any model
        has diagram or shapeCatalog attributes that are not None, Error is
        raised.
    """

    for dexpi_model in models:
        for i in [dexpi_model.diagram, dexpi_model.shapeCatalogues]:
            if i not in [None, []]:
                msg = "Dexpi toolkit does not yet support manipulating diagram information."
                raise NotImplementedError(msg)

    # Retrieve and combine list attributes
    new_model_args = kwargs
    for attr_name, typ in get_type_hints(ConceptualModel).items():
        if get_origin(typ) is list:
            attrs = [getattr(model.conceptualModel, attr_name) for model in models]
            new_model_args[attr_name] = [
                item for model_attr_list in attrs for item in model_attr_list
            ]
        else:
            pass

    new_conceptual_model = ConceptualModel(**new_model_args)

    return DexpiModel(conceptualModel=new_conceptual_model, **new_model_args)


def import_model_contents_into_model(
    target_model: DexpiModel, import_models: list[DexpiModel]
) -> None:
    """Imports the contents of passed models into target model. Conceptually
    similar to combine_dexpi_models, but the target_model is manipulated in
    place and its other, non-list attributes are preserved.

    Parameters
    ----------
    target_model : DexpiModel
        The model that is manipulated in place and that the other models are
        imported into.
    import_models : list[DexpiModel]
        The models that are to be imported to the target.

    Raises
    ------
    NotImplementedError
        Combining diagram information is not yet implemented. If any model
        has diagram or shapeCatalog attributes that are not None, Error is
        raised.
    """
    for dexpi_model in import_models:
        for i in [dexpi_model.diagram, dexpi_model.shapeCatalogues]:
            if i not in [None, []]:
                msg = "Dexpi toolkit does not yet support manipulating diagram information."
                raise NotImplementedError(msg)

    for attr_name, typ in get_type_hints(ConceptualModel).items():
        if get_origin(typ) is list:
            attrs = [getattr(model.conceptualModel, attr_name) for model in import_models]
            new_attrs = [item for model_attr_list in attrs for item in model_attr_list]

            # Append new attributes
            getattr(target_model.conceptualModel, attr_name).extend(new_attrs)


def get_all_instances_in_model(
    the_model: DexpiBaseModel, dexpi_classes: tuple[DexpiBaseModel] | None = None
) -> list[DexpiBaseModel]:
    """
    Recursively discover and collect all instances of specified classes within a model.

    This function traverses the attributes of the provided model object, recursively inspecting
    its attributes and sub-attributes to find all instances of the specified classes. The function
    returns a list of all discovered instances.

    Parameters
    ----------
    the_model : DexpiBaseModel
        The root model object to inspect.
    dexpi_classes : class or tuple of classes, optional
        A tuple of class types to discover within the model. Only instances of these classes will be collected.
        If None, all objects are collected.

    Returns
    -------
    list
        A list of all discovered instances of the specified classes within the model.
    """

    def discover_instances(obj, discovered_instances: tuple) -> list[DexpiBaseModel]:
        """
        Recursively get all compositional attributes of a Python object.

        Parameters
        ----------
        obj : object
            The object to inspect.
        discovered_instances : tuple or any
            The set that keeps track of the discovered instances.

        Returns
        -------
        discovered_instances : list
            Returns the set with discovered instances.
        """

        if dexpi_classes is None:
            if obj not in discovered_instances:
                discovered_instances.append(obj)
        elif isinstance(obj, dexpi_classes) and obj not in discovered_instances:
            discovered_instances.append(obj)

        for attr_name in obj.__class__.model_fields:
            # Skip attributes that compositional
            attr_schema = obj.__class__.model_fields[attr_name].json_schema_extra
            if attr_schema is not None:
                attr_type = attr_schema["attribute_category"]
                if attr_type != "composition":
                    continue
            attr_value = getattr(obj, attr_name)

            if isinstance(attr_value, DexpiBaseModel):
                discovered_instances = discover_instances(attr_value, discovered_instances)
            elif isinstance(attr_value, list):
                for element in attr_value:
                    discovered_instances = discover_instances(element, discovered_instances)

        return discovered_instances

    discovered_instances = []
    return discover_instances(the_model, discovered_instances)


def get_instances_with_condition(
    the_model: DexpiBaseModel,
    condition: Callable[[DexpiBaseModel], bool],
    dexpi_classes: tuple[DexpiBaseModel] | None = None,
) -> list[DexpiBaseModel]:
    """Retrieve all subinstances in a dexpi model that satisfy a specific condition.

    Recursively searches through the model's attributes and sub-attributes to find all instances
    that satisfy the given condition. The condition is specified as a callable that takes a
    DexpiBaseModel instance as an argument and returns a boolean value.

    Parameters
    ----------
    the_model : DexpiBaseModel
        The model to be searched.
    condition : Callable[[DexpiBaseModel], bool]
        The condition that the instances must satisfy.
    dexpi_classes : tuple, optional
        Can specify the classes if only certain class types are to be
        considered. If None, all are considered, by default None.

    Returns
    -------
    list
        List of discovered instances that satisfy the condition.
    """
    instances_with_condition = []

    # Discover all candidates in the model of the candidate classes
    all_candidates = get_all_instances_in_model(the_model, dexpi_classes)

    for candidate in all_candidates:
        if condition(candidate):
            instances_with_condition.append(candidate)

    return instances_with_condition


def get_instances_with_attribute(
    the_model: DexpiBaseModel,
    attribute_name: str,
    target_value: Any = object(),
    dexpi_classes: tuple[DexpiBaseModel] | None = None,
) -> list[DexpiBaseModel]:
    """Retrieve all subinstances in a dexpi model with a specific attribute.

    The dexpi model is searched recursively through compositional attributes
    of all levels. The attribute name is specified as a string. If the target_value is not
    specified, any value is permitted. If the target_value is specified, only instances with
    that value are returned. The dexpi_classes parameter can be used to specify the classes
    to be considered. If None, all classes are considered.

    Parameters
    ----------
    the_model : DexpiBaseModel
        The model to be searched.
    attribute_name : str
        The name of the attribute to be identified.
    target_value : Any, optional
        The value this attribute is supposed to have. If not specified any value is permitted.
    dexpi_classes : tuple, optional
        Can specify the classes if only certain class types are to be
        considered. If None, all are considered, by default None.

    Returns
    -------
    list
        List of discovered instances that have the attribute specified.
    """

    def has_matching_attribute(candidate: DexpiBaseModel) -> bool:
        # Try retrieving the attribute
        try:
            actual_value = getattr(candidate, attribute_name)
            return actual_value == target_value or type(target_value) is object
        except AttributeError:
            # If none found, look in the custom attributes
            if isinstance(candidate, CustomAttributeOwner):
                for custom_attribute in candidate.customAttributes:
                    if custom_attribute.attributeName == attribute_name:
                        actual_value = custom_attribute.value
                        return actual_value == target_value or type(target_value) is object
            return False

    instances_with_attribute = get_instances_with_condition(
        the_model, has_matching_attribute, dexpi_classes=dexpi_classes
    )

    return instances_with_attribute


def object_is_in_model(dexpi_model: DexpiModel, dexpi_object: DexpiBaseModel) -> bool:
    """Checks if the given dexpi_object is in some way a member of the given
    dexpi_model.

    Parameters
    ----------
    dexpi_model : DexpiModel
        The model to be searched for the object.
    dexpi_object : DexpiBaseModel
        The object to be searched for in the model.

    Returns
    -------
    bool
        True if object is found in the model, False otherwise
    """
    candidates = get_all_instances_in_model(dexpi_model, type(dexpi_object))
    if dexpi_object in candidates:
        return True
    else:
        return False



"""Contains the factory for instantiating parser objects for XML elements."""

from collections.abc import Callable
from xml.etree import ElementTree as ET

from pydexpi.loaders.proteus_serializer.core import ModuleContext, ParserModule
from pydexpi.loaders.proteus_serializer.parser_modules import (
    ActuatingElectricalFunctionParser,
    ActuatingFunctionParser,
    ActuatingSystemComponentParser,
    ActuatingSystemParser,
    AssociationParser,
    CenterLineParser,
    ConnectionPointParser,
    EquipmentParser,
    GenericAttributeParser,
    InformationFlowParser,
    InstrumentationLoopFunctionParser,
    NozzleParser,
    OffPageConnectorReferenceParser,
    PipeOffPageConnectorParser,
    PipingComponentParser,
    PipingNetworkSegmentParser,
    PipingNetworkSystemParser,
    PipingNodeParser,
    PlantModelParser,
    ProcessInstrumentationFunctionParser,
    ProcessSignalGeneratingFunctionParser,
    ProcessSignalGeneratingSystemComponentParser,
    ProcessSignalGeneratingSystemParser,
    PropertyBreakParser,
    SignalOffPageConnectorParser,
)


class ParserFactory:
    """Factory class that creates parsers given an XML element.

    Includes a factory method for each xml element type and corresponding parser class. If not
    otherwise specified, the parser is created with the XML element and context as parameters, and
    returns an instance of the corresponding parser class. The top level method is
    `make_plant_model_parser`, which creates a parser for the top level plant model element.
    """

    def __init__(self) -> None:
        """Initialize the ParserFactory by creating the factory method registry for the instance."""
        self.factory_methods: dict[str, Callable[[ModuleContext, ET.Element], ParserModule]] = {
            "PlantModel": self.make_plant_model_parser,
            "ActuatingElectricalFunction": self.make_actuating_electrical_function_parser,
            "ActuatingFunction": self.make_actuating_function_parser,
            "ActuatingSystemComponent": self.make_actuating_system_component_parser,
            "ActuatingSystem": self.make_actuating_system_parser,
            "Association": self.make_association_parser,
            "CenterLine": self.make_center_line_parser,
            "ConnectionPoints": self.make_connection_point_parser,
            "Equipment": self.make_equipment_parser,
            "GenericAttribute": self.make_generic_attribute_parser,
            "InformationFlow": self.make_information_flow_parser,
            "InformationFlowOffPageConnectorReference": self.make_opc_reference_parser,
            "InstrumentationLoopFunction": self.make_instrumentation_loop_function_parser,
            "Nozzle": self.make_nozzle_parser,
            "PipeOffPageConnector": self.make_pipe_off_page_connector_parser,
            "PipeOffPageConnectorReference": self.make_opc_reference_parser,
            "PipingComponent": self.make_piping_component_parser,
            "PipingNetworkSegment": self.make_piping_network_segment_parser,
            "PipingNetworkSystem": self.make_piping_network_system_parser,
            "PipingNode": self.make_piping_node_parser,
            "ProcessInstrumentationFunction": self.make_process_instrumentation_function_parser,
            "ProcessSignalGeneratingFunction": (
                self.make_process_signal_generating_function_parser
            ),
            "ProcessSignalGeneratingSystemComponent": (
                self.make_process_signal_generating_system_component_parser
            ),
            "ProcessSignalGeneratingSystem": (self.make_process_signal_generating_system_parser),
            "PropertyBreak": self.make_property_break_parser,
            "InformationFlowOffPageConnector": self.make_signal_off_page_connector_parser,
        }

    def _parse_child_elements(
        self,
        parent_element: ET.Element,
        context: ModuleContext,
        element_tag: str,
        factory_method_key: str = None,
    ) -> list[ParserModule]:
        """Helper method to parse child elements of a given tag using the provided factory method.

        Parameters
        ----------
        parent_element : ET.Element
            The parent XML element containing child elements to parse.
        context : ModuleContext
            The context in which the XML element is being parsed.
        element_tag : str
            The tag name of the child elements to parse.
        factory_method_key : str
            The key corresponding to the factory method to use for creating parsers. If None, the
            element tag is used as the key.

        Returns
        -------
        list
            A list of parser instances created for the child elements.
        """
        if factory_method_key is None:
            factory_method_key = element_tag

        parsers = []
        for child in parent_element.findall(element_tag):
            sub_context = context.get_updated_context(child)
            parser = self.factory_methods[factory_method_key](sub_context, child)
            parsers.append(parser)
        return parsers

    def _create_parser_with_common_elements(
        self,
        parser_class: type,
        context: ModuleContext,
        element: ET.Element,
        parse_associations: bool = True,
        parse_generic_attributes: bool = True,
        **additional_parsers,
    ) -> ParserModule:
        """Helper to create parsers with common elements.

        Common elements include associations and generic attributes."""
        kwargs = {
            "context": context,
            "element": element,
        }

        if parse_associations:
            kwargs["association_parsers"] = self._parse_child_elements(
                element, context, "Association"
            )

        if parse_generic_attributes:
            kwargs["generic_attribute_parser"] = self.make_generic_attribute_parser(
                context, element
            )

        kwargs.update(additional_parsers)
        return parser_class(**kwargs)

    ### GENERIC MODULES ###
    def make_generic_attribute_parser(
        self, context: ModuleContext, parent_element: ET.Element
    ) -> GenericAttributeParser:
        """Creates a generic attribute parser for the given XML element.

        Note that the element is not the generic attribute element itself, but the parent
        element that contains the generic attributes.
        """
        return GenericAttributeParser(parent_element=parent_element, context=context)

    def make_association_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> AssociationParser:
        """Creates an association parser for the given XML element."""
        return AssociationParser(element=element, context=context)

    def make_opc_reference_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> OffPageConnectorReferenceParser:
        """Creates an off-page connector parser for the given XML element.

        Includes parsing associations and generic attributes.
        """
        opc_reference_parser = self._create_parser_with_common_elements(
            OffPageConnectorReferenceParser, context, element
        )

        return opc_reference_parser

    ### EQUIPMENT MODULES ###
    def make_nozzle_parser(self, context: ModuleContext, element: ET.Element) -> NozzleParser:
        """
        Creates a nozzle parser for the given XML element.

        Includes parsing connection points, associations, and generic attributes.
        """
        # Collect all connection point parsers defined in the nozzle
        connection_point_parsers = self._parse_child_elements(element, context, "ConnectionPoints")

        # Create and return the NozzleParser instance
        nozzle_parser = self._create_parser_with_common_elements(
            NozzleParser,
            context,
            element,
            connection_point_parsers=connection_point_parsers,
        )
        return nozzle_parser

    def make_equipment_parser(self, context: ModuleContext, element: ET.Element) -> EquipmentParser:
        """Creates an equipment parser for the given XML element.

        Includes parsing subequipment, nozzles, associations, and generic attributes.
        """
        # Make subequipment parsers
        subequipment_parsers = self._parse_child_elements(element, context, "Equipment")

        # Make nozzle parsers
        nozzle_parsers = self._parse_child_elements(element, context, "Nozzle")

        # Create and return the EquipmentParser instance
        equipment_parser = self._create_parser_with_common_elements(
            EquipmentParser,
            context,
            element,
            subequipment_parsers=subequipment_parsers,
            nozzle_parsers=nozzle_parsers,
        )
        return equipment_parser

    ### PIPING MODULES ###
    def make_piping_node_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> PipingNodeParser:
        """Creates a piping node parser for the given XML element with generic attributes."""

        # Create and return the PipingNodeParser instance
        node_parser = self._create_parser_with_common_elements(
            PipingNodeParser,
            context,
            element,
            parse_associations=False,
        )
        return node_parser

    def make_connection_point_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> ConnectionPointParser:
        """Creates a connection point parser for the given XML element.

        Includes parsing piping nodes, which are nodes of type "process".
        """
        # Collect all piping node parsers defined in the connection point
        # These are nodes of type "process"
        piping_node_parsers = []
        for node in element.findall("Node"):
            if node.get("Type") == "process":
                sub_context = context.get_updated_context(node)
                piping_node_parsers.append(self.make_piping_node_parser(sub_context, node))

        # Create and return the ConnectionPointParser instance
        connection_point_parser = ConnectionPointParser(
            context=context,
            element=element,
            piping_node_parsers=piping_node_parsers,
        )
        return connection_point_parser

    def make_piping_component_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> PipingComponentParser:
        """Creates a piping component parser for the given XML element.

        Includes parsing connection points, associations, and generic attributes.
        """
        # Collect all connection point parsers defined in the piping component
        connection_point_parsers = self._parse_child_elements(element, context, "ConnectionPoints")

        parser = self._create_parser_with_common_elements(
            PipingComponentParser,
            context,
            element,
            connection_point_parsers=connection_point_parsers,
        )
        return parser

    def make_pipe_off_page_connector_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> PipeOffPageConnectorParser:
        """Creates an off-page connector parser for the given XML element.

        Includes parsing connection points, associations, and generic attributes.
        """
        # Collect all connection point parsers defined in the off-page connector
        connection_point_parsers = self._parse_child_elements(element, context, "ConnectionPoints")

        # Make reference parsers
        reference_parsers = self._parse_child_elements(
            element, context, "PipeOffPageConnectorReference"
        )

        parser = self._create_parser_with_common_elements(
            PipeOffPageConnectorParser,
            context,
            element,
            connection_point_parsers=connection_point_parsers,
            reference_parsers=reference_parsers,
        )

        return parser

    def make_property_break_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> PropertyBreakParser:
        """Creates a property break parser for the given XML element.

        Includes parsing connection points, associations, and generic attributes."""
        # Collect all connection point parsers defined in the property break
        connection_point_parsers = self._parse_child_elements(element, context, "ConnectionPoints")

        parser = self._create_parser_with_common_elements(
            PropertyBreakParser,
            context,
            element,
            connection_point_parsers=connection_point_parsers,
        )

        return parser

    def make_center_line_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> CenterLineParser:
        """Creates a center line parser for the given XML element with generic attributes."""
        return self._create_parser_with_common_elements(
            CenterLineParser,
            context,
            element,
            parse_associations=False,
        )

    def make_piping_network_segment_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> PipingNetworkSegmentParser:
        """
        Creates a piping network segment parser for the given XML element.

        Makes parsers for components, OPCs, property breaks, and center lines. Keep track of
        component, OPC, and property break parsers as items and center line parsers as
        connections. Also keep track of all parsers in order, as this is important to infer
        connectivity.
        """

        # Make parsers for components, OPCs, property breaks, and center lines. Keep track of
        # component, OPC, and property break parsers as items and center line parsers as
        # connections. Also keep track of all parsers in order, as this is important to infer
        # connectivity.
        ordered_element_parsers = []
        item_parsers = []
        center_line_parsers = []
        for subelement in element:
            if subelement.tag == "PipingComponent":
                sub_context = context.get_updated_context(subelement)
                component_parser = self.make_piping_component_parser(sub_context, subelement)
                ordered_element_parsers.append(component_parser)
                item_parsers.append(component_parser)
            elif subelement.tag == "PipeOffPageConnector":
                sub_context = context.get_updated_context(subelement)
                component_parser = self.make_pipe_off_page_connector_parser(sub_context, subelement)
                ordered_element_parsers.append(component_parser)
                item_parsers.append(component_parser)
            elif subelement.tag == "PropertyBreak":
                sub_context = context.get_updated_context(subelement)
                component_parser = self.make_property_break_parser(sub_context, subelement)
                ordered_element_parsers.append(component_parser)
                item_parsers.append(component_parser)
            elif subelement.tag == "CenterLine":
                sub_context = context.get_updated_context(subelement)
                component_parser = self.make_center_line_parser(sub_context, subelement)
                ordered_element_parsers.append(component_parser)
                center_line_parsers.append(component_parser)

        parser = self._create_parser_with_common_elements(
            PipingNetworkSegmentParser,
            context,
            element,
            ordered_element_parsers=ordered_element_parsers,
            item_parsers=item_parsers,
            center_line_parsers=center_line_parsers,
        )

        return parser

    def make_piping_network_system_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> PipingNetworkSystemParser:
        """Creates a piping network system parser for the given XML element.

        Includes parsing segments and generic attributes.
        """
        # Make segment parsers
        segment_parsers = self._parse_child_elements(element, context, "PipingNetworkSegment")

        # Create and return the PipingNetworkSystemParser instance
        parser = self._create_parser_with_common_elements(
            PipingNetworkSystemParser,
            context,
            element,
            segment_parsers=segment_parsers,
            parse_associations=False,
        )

        return parser

    ### Instrumentation modules ###
    def make_actuating_system_component_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> ActuatingSystemComponentParser:
        """Creates an actuating system component parser for the given XML element.

        Includes parsing associations and generic attributes.
        """
        return self._create_parser_with_common_elements(
            ActuatingSystemComponentParser, context, element
        )

    def make_actuating_system_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> ActuatingSystemParser:
        """Creates an actuating system parser for the given XML element.

        Includes parsing components and generic attributes.
        """
        # Make component parsers
        component_parsers = self._parse_child_elements(element, context, "ActuatingSystemComponent")

        # Create and return the ActuatingSystemParser instance
        parser = self._create_parser_with_common_elements(
            ActuatingSystemParser,
            context,
            element,
            component_parsers=component_parsers,
        )

        return parser

    def make_process_signal_generating_system_component_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> ProcessSignalGeneratingSystemComponentParser:
        """Creates a process signal generating system component parser for the given XML element.

        Includes parsing associations and generic attributes.
        """
        return self._create_parser_with_common_elements(
            ProcessSignalGeneratingSystemComponentParser, context, element
        )

    def make_process_signal_generating_system_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> ProcessSignalGeneratingSystemParser:
        """Creates a process signal generating system parser for the given XML element.

        Includes parsing components and generic attributes.
        """
        # Make component parsers
        component_parsers = self._parse_child_elements(
            element, context, "ProcessSignalGeneratingSystemComponent"
        )

        # Create and return the ProcessSignalGeneratingSystemParser instance
        parser = self._create_parser_with_common_elements(
            ProcessSignalGeneratingSystemParser,
            context,
            element,
            component_parsers=component_parsers,
            parse_associations=False,
        )

        return parser

    def make_actuating_function_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> ActuatingFunctionParser:
        """Creates an actuating function parser for the given XML element.

        Includes parsing for generic attributes and associations.
        """
        return self._create_parser_with_common_elements(ActuatingFunctionParser, context, element)

    def make_actuating_electrical_function_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> ActuatingElectricalFunctionParser:
        """Creates an actuating electrical function parser for the given XML element.

        Includes parsing for generic attributes and associations.
        """
        return self._create_parser_with_common_elements(
            ActuatingElectricalFunctionParser, context, element
        )

    def make_process_signal_generating_function_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> ProcessSignalGeneratingFunctionParser:
        """Creates a process signal generating function parser for the given XML element.

        Includes parsing for generic attributes and associations.
        """
        return self._create_parser_with_common_elements(
            ProcessSignalGeneratingFunctionParser, context, element
        )

    def make_signal_off_page_connector_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> SignalOffPageConnectorParser:
        """Creates a signal off-page connector parser for the given XML element.

        Includes parsing references and associations.
        """
        # Make reference parsers
        reference_parsers = self._parse_child_elements(
            element, context, "InformationFlowOffPageConnectorReference"
        )

        parser = self._create_parser_with_common_elements(
            SignalOffPageConnectorParser,
            context,
            element,
            reference_parsers=reference_parsers,
            parse_generic_attributes=False,
        )

        return parser

    def make_information_flow_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> InformationFlowParser:
        """Creates an information flow parser for the given XML element.

        Includes parsing for generic attributes and associations.
        """
        return self._create_parser_with_common_elements(InformationFlowParser, context, element)

    def make_process_instrumentation_function_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> ProcessInstrumentationFunctionParser:
        """Creates a process instrumentation function parser for the given XML element.

        Includes parsing for actuating electrical functions, actuating functions, signal generating
        functions, signal opcs, signal conveying functions, generic attributes, and associations.
        """
        # Make actuating electrical function parsers
        actuating_electrical_function_parsers = self._parse_child_elements(
            element, context, "ActuatingElectricalFunction"
        )

        # Make actuating function parsers
        actuating_function_parsers = self._parse_child_elements(
            element, context, "ActuatingFunction"
        )

        # Make process signal generating function parsers
        process_signal_generating_function_parsers = self._parse_child_elements(
            element, context, "ProcessSignalGeneratingFunction"
        )

        # Make signal off-page connector parsers
        signal_off_page_connector_parsers = self._parse_child_elements(
            element, context, "InformationFlowOffPageConnector"
        )

        # Make information flow parsers
        information_flow_parsers = self._parse_child_elements(element, context, "InformationFlow")

        parser = self._create_parser_with_common_elements(
            ProcessInstrumentationFunctionParser,
            context,
            element,
            actuating_electrical_function_parsers=actuating_electrical_function_parsers,
            actuating_function_parsers=actuating_function_parsers,
            signal_generating_function_parsers=process_signal_generating_function_parsers,
            signal_opc_parsers=signal_off_page_connector_parsers,
            information_flow_parsers=information_flow_parsers,
        )

        # Create and return the ProcessInstrumentationFunctionParser instance
        return parser

    def make_instrumentation_loop_function_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> InstrumentationLoopFunctionParser:
        """Creates an instrumentation loop function parser for the given XML element.

        Includes parsing associations and generic attributes.
        """
        return self._create_parser_with_common_elements(
            InstrumentationLoopFunctionParser, context, element
        )

    ### MODEL PARSERS ###
    def make_plant_model_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> PlantModelParser:
        """Creates a plant model parser for the given XML element.

        Includes parsing equipment, piping network systems, and actuating systems.
        """
        # Make equipment parsers
        equipment_parsers = self._parse_child_elements(element, context, "Equipment")

        # Make piping network system parsers
        piping_network_system_parsers = self._parse_child_elements(
            element, context, "PipingNetworkSystem"
        )

        # Make actuating system parsers
        actuating_system_parsers = self._parse_child_elements(element, context, "ActuatingSystem")

        # Make process signal generating system parsers
        process_signal_generating_system_parsers = self._parse_child_elements(
            element, context, "ProcessSignalGeneratingSystem"
        )

        # Make process instrumentation function parsers
        process_instrumentation_function_parsers = self._parse_child_elements(
            element, context, "ProcessInstrumentationFunction"
        )

        # Make instrumentation loop function parsers
        instrumentation_loop_function_parsers = self._parse_child_elements(
            element, context, "InstrumentationLoopFunction"
        )

        return PlantModelParser(
            context=context,
            element=element,
            equipment_parsers=equipment_parsers,
            piping_network_system_parsers=piping_network_system_parsers,
            actuating_system_parsers=actuating_system_parsers,
            process_signal_generating_system_parsers=process_signal_generating_system_parsers,
            process_instrumentation_function_parsers=process_instrumentation_function_parsers,
            instrumentation_loop_function_parsers=instrumentation_loop_function_parsers,
        )



"""Tests for GeneratorFunction on a basic TestGeneratorFunction and RandomGeneratorFunction."""

import random

import pytest

from pydexpi.syndata.generation_history import GenerationHistory
from pydexpi.syndata.generator_function import (
    CappingFunction,
    GeneratorFunction,
    RandomGeneratorFunction,
    ReconstructionCappingFunction,
    ReconstructionGeneratorFunction,
)
from pydexpi.syndata.generator_step import (
    AddPatternStep,
    CappingStep,
    GeneratorStep,
    InitializationStep,
    InternalConnectionStep,
    TerminationStep,
)
from pydexpi.syndata.pattern import Connector, Pattern


class TestGeneratorFunction(GeneratorFunction):
    """Test class for GeneratorFunction.

    This class is used to test the GeneratorFunction class. It is a subclass of
    GeneratorFunction and implements the abstract methods.
    """

    __test__ = False

    def __init__(self, distributions: dict, test_connect_internal: bool = False) -> None:
        """
        Initialize the TestGeneratorFunction.
        """
        super().__init__(distributions)
        self.test_connect_internal = test_connect_internal

    def initialize_pattern(self) -> InitializationStep:
        """
        Initialize a pattern using the first pattern from the "Distr1" distribution.
        """
        the_distribution = self.distribution_range["Distr1"]
        the_pattern = list(the_distribution.patterns.values())[0]
        return InitializationStep(the_pattern, the_distribution.name)

    def get_next_step(self, pattern: Pattern) -> GeneratorStep:
        """
        Get the next step for the given pattern.
        """
        if self.test_connect_internal:
            # If test_connect_internal is True, create an InternalConnectionStep
            return InternalConnectionStep(
                list(pattern.connectors.values())[0], list(pattern.connectors.values())[1]
            )
        else:
            the_distribution = self.distribution_range["Distr1"]
            next_pattern = list(the_distribution.patterns.values())[0]
            return AddPatternStep(
                list(pattern.connectors.values())[0],
                next_pattern,
                list(next_pattern.connectors.values())[0],
                the_distribution.name,
            )


def test_testgeneratorfunction_initialize_pattern(simple_distribution_factory):
    """Test that TestGeneratorFunction returns a valid InitializationStep."""
    distributions = {name: simple_distribution_factory(name) for name in ["Distr1", "Distr2"]}
    test_func = TestGeneratorFunction(distributions)
    init_step = test_func.initialize_pattern()
    the_pattern = init_step.get_pattern()
    assert isinstance(the_pattern, Pattern)


def test_testgeneratorfunction_get_next_step_add(
    simple_distribution_factory, simple_pattern_factory
):
    """Test get_next_step returns AddPatternStep when test_connect_internal is False."""
    distributions = {name: simple_distribution_factory(name) for name in ["Distr1", "Distr2"]}
    test_func = TestGeneratorFunction(distributions, test_connect_internal=False)
    pattern = simple_pattern_factory("TestPattern")
    step = test_func.get_next_step(pattern)
    assert step.__class__.__name__ == "AddPatternStep"
    assert step.own_connector in pattern.connectors.values()
    assert isinstance(step.next_pattern, Pattern)
    assert step.next_connector in step.next_pattern.connectors.values()


def test_testgeneratorfunction_get_next_step_internal(
    simple_distribution_factory, simple_pattern_factory
):
    """Test get_next_step returns InternalConnectionStep when test_connect_internal is True."""
    distributions = {name: simple_distribution_factory(name) for name in ["Distr1", "Distr2"]}
    test_func = TestGeneratorFunction(distributions, test_connect_internal=True)
    pattern = simple_pattern_factory("TestPattern")
    step = test_func.get_next_step(pattern)
    assert step.__class__.__name__ == "InternalConnectionStep"
    assert step.own_connector in pattern.connectors.values()
    # Since next pattern is selected from distribution, validate that next_connector exists
    assert isinstance(step.next_connector, Connector)


def test_random_generator_function_constructor(simple_distribution_factory):
    """Test construction of a generator function by instantiating a RandomGeneratorFunction."""
    distributions = {name: simple_distribution_factory(name) for name in ["Distr1", "Distr2"]}
    the_function = RandomGeneratorFunction(distributions)
    assert isinstance(the_function, GeneratorFunction)
    assert isinstance(the_function, RandomGeneratorFunction)


def test_random_function_initialize_pattern(simple_distribution_factory):
    """Test initializing a pattern with the simple example
    RandomGeneratorFunction."""
    random.seed(42)
    distributions = {name: simple_distribution_factory(name) for name in ["Distr1", "Distr2"]}
    the_function = RandomGeneratorFunction(distributions)
    initialization_step = the_function.initialize_pattern()
    # get the initialized pattern from the step
    the_pattern = initialization_step.get_pattern()
    assert isinstance(the_pattern, Pattern)


def test_random_generator_function(simple_distribution_factory, simple_pattern_factory):
    """Test getting the next step from RandomGeneratorFunction with a simple pattern."""
    random.seed(42)
    the_pattern = simple_pattern_factory("Some label")

    # Case: set p_connect_internal to 0 (expect an AddPatternStep)
    distributions = {name: simple_distribution_factory(name) for name in ["Distr1", "Distr2"]}
    the_function = RandomGeneratorFunction(distributions)

    step = the_function.get_next_step(the_pattern)
    assert isinstance(step, AddPatternStep)

    selected_connector = step.own_connector
    next_pattern = step.next_pattern
    next_connector = step.next_connector

    assert isinstance(selected_connector, Connector)
    assert isinstance(next_connector, Connector)
    assert isinstance(next_pattern, Pattern)
    assert selected_connector in the_pattern.connectors.values()
    assert next_connector in next_pattern.connectors.values()

    # Case: set p_connect_internal to 1 (expect an InternalConnectionStep)
    the_function.set_p_connect_internal(1)
    step = the_function.get_next_step(the_pattern)
    assert isinstance(step, InternalConnectionStep)

    selected_connector = step.own_connector
    next_connector = step.next_connector

    assert isinstance(selected_connector, Connector)
    assert isinstance(next_connector, Connector)
    assert selected_connector in the_pattern.connectors.values()
    assert next_connector in the_pattern.connectors.values()


def test_capping_function_constructor():
    """ "Test construction of a generator function by instantiating a CappingFunction."""
    my_function = CappingFunction()
    assert isinstance(my_function, CappingFunction)


def test_capping_function_get_capping_steps(simple_pattern_factory):
    """Test getting capping steps from CappingFunction."""
    my_function = CappingFunction()

    pattern1 = simple_pattern_factory("dummy")

    capping_steps = my_function.get_capping_steps(pattern1)

    assert isinstance(capping_steps, list)
    assert len(capping_steps) == 2
    assert isinstance(capping_steps[0], CappingStep)

    # Make sure capping steps are valid by executing it
    for step in capping_steps:
        step.execute_on(pattern1)

    # Corrupt internal _make_capping_step method to test error handling
    pattern1 = simple_pattern_factory("dummy")
    my_function._make_capping_steps = lambda x: []
    with pytest.raises(RuntimeError, match="Connectors of the CappingSteps do not match"):
        my_function.get_capping_steps(pattern1)


def test_reconstruction_generator_function_constructor(simple_distribution_factory):
    """Test construction of a reconstruction generator function."""
    distributions = {name: simple_distribution_factory(name) for name in ["Distr1", "Distr2"]}
    the_function = ReconstructionGeneratorFunction(distributions)
    assert isinstance(the_function, GeneratorFunction)
    assert isinstance(the_function, ReconstructionGeneratorFunction)
    assert the_function.generation_history is None
    assert the_function._current_step == 0


def test_reconstruction_generator_function_set_history():
    """Test setting the generation history."""
    reconstruction_func = ReconstructionGeneratorFunction({})
    history = GenerationHistory()
    reconstruction_func.set_generation_history(history)
    assert reconstruction_func.generation_history is history
    assert reconstruction_func._current_step == 0


def test_reconstruction_generator_function_no_history(simple_distribution_factory):
    """Test error handling when generation history is not set."""
    distributions = {name: simple_distribution_factory(name) for name in ["Distr1", "Distr2"]}
    the_function = ReconstructionGeneratorFunction(distributions)

    with pytest.raises(AttributeError, match="Generation history is not set"):
        the_function.initialize_pattern()

    # Create a dummy pattern to test get_next_step
    pattern = next(iter(distributions.values())).patterns[
        next(iter(distributions["Distr1"].patterns.keys()))
    ]

    with pytest.raises(AttributeError, match="Generation history is not set"):
        the_function.get_next_step(pattern)


def test_reconstruction_initialize_pattern(simple_distribution_factory):
    """Test initializing a pattern with a valid history."""
    distributions = {name: simple_distribution_factory(name) for name in ["Distr1", "Distr2"]}
    history = GenerationHistory()

    # Add a mock initialization step to history
    history.history.append(
        {
            "generator_step_type": "initialization",
            "own_connector": None,
            "next_pattern": "pattern1",
            "next_connector": None,
            "sampled_distribution_name": "Distr1",
        }
    )

    the_function = ReconstructionGeneratorFunction(distributions, history)

    # Test initialize_pattern
    init_step = the_function.initialize_pattern()
    assert isinstance(init_step, InitializationStep)
    assert init_step.sampled_distribution_name == "Distr1"
    assert the_function._current_step == 1


def test_reconstruction_get_next_step(simple_distribution_factory):
    """Test getting next steps with a valid history."""
    distributions = {name: simple_distribution_factory(name) for name in ["Distr1", "Distr2"]}
    history = GenerationHistory()

    # Get actual patterns from the distributions to work with
    pattern1 = next(iter(distributions["Distr1"].patterns.values()))
    pattern2 = next(iter(distributions["Distr2"].patterns.values()))

    # Get connector ids from the actual patterns
    connector1 = next(iter(pattern1.connectors.values()))
    connector2 = next(iter(pattern2.connectors.values()))
    connector3 = (
        list(pattern1.connectors.values())[1] if len(pattern1.connectors) > 1 else connector1
    )

    # Setup mock history with different step types
    history.history = [
        {
            "generator_step_type": "initialization",
            "own_connector": None,
            "next_pattern": pattern1.label,
            "next_connector": None,
            "sampled_distribution_name": "Distr1",
        },
        {
            "generator_step_type": "add_pattern",
            "own_connector": connector1.label,
            "next_pattern": pattern2.label,
            "next_connector": connector2.label,
            "sampled_distribution_name": "Distr2",
        },
        {
            "generator_step_type": "internal_connection",
            "own_connector": connector1.label,
            "next_connector": connector3.label,
            "next_pattern": None,
            "sampled_distribution_name": None,
        },
        {
            "generator_step_type": "termination",
            "own_connector": None,
            "next_pattern": None,
            "next_connector": None,
            "sampled_distribution_name": None,
        },
    ]

    the_function = ReconstructionGeneratorFunction(distributions, history)

    # Skip initialization step
    the_function._current_step = 1

    # Test add_pattern step
    next_step = the_function.get_next_step(pattern1)
    assert isinstance(next_step, AddPatternStep)
    assert next_step.own_connector.label == connector1.label
    assert next_step.next_pattern.label == pattern2.label
    assert the_function._current_step == 2

    # Test internal_connection step
    next_step = the_function.get_next_step(pattern1)
    assert isinstance(next_step, InternalConnectionStep)
    assert next_step.own_connector.label == connector1.label
    assert next_step.next_connector.label == connector3.label
    assert the_function._current_step == 3

    # Test termination step
    next_step = the_function.get_next_step(pattern1)
    assert isinstance(next_step, TerminationStep)
    assert the_function._current_step == 4

    # Test Error: Attempting to get next step after all steps have been processed
    invalid_history = GenerationHistory()
    invalid_history.history = history.history.copy()

    # Create a new function with the same history
    error_function = ReconstructionGeneratorFunction(distributions, invalid_history)
    error_function._current_step = 4  # Set to after all steps

    with pytest.raises(IndexError):
        error_function.get_next_step(pattern1)

    # Test Error: Invalid step type in history
    invalid_history = GenerationHistory()
    invalid_history.history = [
        {
            "generator_step_type": "initialization",
            "own_connector": None,
            "next_pattern": pattern1.label,
            "next_connector": None,
            "sampled_distribution_name": "Distr1",
        },
        {
            "generator_step_type": "invalid_step_type",  # Invalid step type
            "own_connector": connector1.label,
            "next_pattern": pattern2.label,
            "next_connector": connector2.label,
            "sampled_distribution_name": "Distr2",
        },
    ]

    error_function = ReconstructionGeneratorFunction(distributions, invalid_history)
    error_function._current_step = 1  # Skip initialization

    with pytest.raises(ValueError, match="Unknown step type"):
        error_function.get_next_step(pattern1)


def test_reconstruction_reset():
    """Test resetting the reconstruction generator function."""
    reconstruction_func = ReconstructionGeneratorFunction({})
    history = GenerationHistory()
    reconstruction_func.set_generation_history(history)

    # Manually set current step to non-zero
    reconstruction_func._current_step = 3

    # Reset should set current step back to 0
    reconstruction_func.reset()
    assert reconstruction_func._current_step == 0


def test_reconstruction_capping_function_constructor(simple_distribution_factory):
    """Test construction of a reconstruction capping function."""
    distributions = {
        name: simple_distribution_factory(name, no_pattern_connectors=1)
        for name in ["Distr1", "Distr2"]
    }
    the_function = ReconstructionCappingFunction(distributions)
    assert isinstance(the_function, CappingFunction)
    assert isinstance(the_function, ReconstructionCappingFunction)
    assert the_function.generation_history is None


def test_reconstruction_capping_function(simple_distribution_factory, simple_pattern_factory):
    """Test making capping steps with a valid history."""
    distributions = {
        name: simple_distribution_factory(name, no_pattern_connectors=1)
        for name in ["Distr1", "Distr2"]
    }
    history = GenerationHistory()

    # Use simple_pattern_factory for the pattern
    pattern = simple_pattern_factory("TestPattern")

    # Get connector labels from the pattern
    conn_labels = list(pattern.connectors.keys())

    # Setup mock capping steps in history
    history.history = [
        {
            "generator_step_type": "capping",
            "own_connector": conn_labels[0],
            "next_pattern": "CapPattern",
            "next_connector": "Connector label0",
            "sampled_distribution_name": "Distr1",
        },
        {
            "generator_step_type": "capping",
            "own_connector": conn_labels[1],
            "next_pattern": None,
            "next_connector": None,
            "sampled_distribution_name": None,
        },
    ]

    # Create the reconstruction capping function with the history
    the_function = ReconstructionCappingFunction(distributions, history)

    # Verify construction
    assert isinstance(the_function, CappingFunction)
    assert isinstance(the_function, ReconstructionCappingFunction)
    assert the_function.generation_history is history

    # Test setting history separately
    new_function = ReconstructionCappingFunction(distributions)
    assert new_function.generation_history is None
    new_function.set_generation_history(history)
    assert new_function.generation_history is history

    # Test Error: Create with distributions having multiple connectors
    with pytest.raises(
        ValueError, match="CappingFunction should only use distributions with one connector label"
    ):
        multi_connector_distributions = {
            name: simple_distribution_factory(name, no_pattern_connectors=2)
            for name in ["Distr1", "Distr2"]
        }
        ReconstructionCappingFunction(multi_connector_distributions)

    # Test Error: History with invalid capping step (missing connector)
    incomplete_history = GenerationHistory()
    # Only include one connector, missing the second
    incomplete_history.history = [history.history[0]]

    error_function = ReconstructionCappingFunction(distributions, incomplete_history)

    # This would fail in _make_capping_steps if we were able to test it directly
    # Since the original test doesn't actually call _make_capping_steps, we're verifying the attribute
    assert error_function.generation_history is incomplete_history



"""Tests for the connector_renaming module."""

from pydexpi.syndata.connector_renaming import ConnectorRenamingConvention


def test_connector_renaming_convention_init():
    """Test initialization of a ConnectorRenamingConvention."""
    renaming = ConnectorRenamingConvention()
    assert renaming.encountered_pattern_counter == {}


def test_connector_renaming_convention_rename_connectors(simple_pattern_factory):
    """Test renaming connectors in a pattern."""
    # Create a pattern
    pattern = simple_pattern_factory("TestPattern")
    # Get initial connector labels
    original_labels = list(pattern.connectors.keys())

    # Create renaming convention
    renaming = ConnectorRenamingConvention()

    # Rename all connectors
    renaming.rename_connectors(pattern, [])

    # Check that connector labels have changed
    new_labels = list(pattern.connectors.keys())
    assert set(original_labels) != set(new_labels)

    # Check format of new labels
    for label in new_labels:
        assert label.startswith("TestPattern_")
        # Should follow format pattern_counter_original_label
        parts = label.split("_")
        assert parts[0] == "TestPattern"
        assert parts[1].isdigit()

        # Label key convention retained
        assert label == pattern.connectors[label].label


def test_connector_renaming_convention_rename_skip_connectors(simple_pattern_factory):
    """Test skipping specific connectors during renaming."""
    # Create a pattern
    pattern = simple_pattern_factory("TestPattern")

    # Get a connector to skip
    connectors_list = list(pattern.connectors.values())
    connector_to_skip = connectors_list[0]
    original_label = connector_to_skip.label

    # Create renaming convention
    renaming = ConnectorRenamingConvention()

    # Rename connectors, skipping one
    renaming.rename_connectors(pattern, [connector_to_skip])

    # Check that skipped connector label didn't change
    assert connector_to_skip.label == original_label

    # Check other connector was renamed
    other_connector = connectors_list[1]
    assert other_connector.label != original_label
    assert other_connector.label.startswith("TestPattern_")


def test_connector_renaming_convention_reset_counter(simple_pattern_factory):
    """Test resetting the counter."""
    # Create patterns
    pattern1 = simple_pattern_factory("TestPattern")
    pattern2 = simple_pattern_factory("TestPattern")

    # Create renaming convention
    renaming = ConnectorRenamingConvention()

    # Rename first pattern connectors
    renaming.rename_connectors(pattern1, [])
    first_renamed_labels = list(pattern1.connectors.keys())

    # Reset counter
    renaming.reset()

    # Rename second pattern connectors
    renaming.rename_connectors(pattern2, [])
    second_renamed_labels = list(pattern2.connectors.keys())

    # The labels should be the same since counter was reset
    assert set(first_renamed_labels) == set(second_renamed_labels)


def test_connector_renaming_convention_incremental_renaming(simple_pattern_factory):
    """Test that pattern counter increments for subsequent patterns with same label."""
    # Create patterns with the same label
    pattern1 = simple_pattern_factory("SameLabel")
    pattern2 = simple_pattern_factory("SameLabel")

    # Create renaming convention
    renaming = ConnectorRenamingConvention()

    # Rename first pattern
    renaming.rename_connectors(pattern1, [])
    pattern1_labels = list(pattern1.connectors.keys())

    # All labels should have index 0
    for label in pattern1_labels:
        parts = label.split("_")
        assert parts[0] == "SameLabel"
        assert parts[1] == "0"

    # Rename second pattern
    renaming.rename_connectors(pattern2, [])
    pattern2_labels = list(pattern2.connectors.keys())

    # All labels should have index 1
    for label in pattern2_labels:
        parts = label.split("_")
        assert parts[0] == "SameLabel"
        assert parts[1] == "1"



"""This module contains utility functions for pydexpi base models. Becasue these
are the common parents to the pydexpi classes, these are basic, overarching
functionalities for all dexpi classes."""

from pydexpi import dexpi_classes
from pydexpi.dexpi_classes.dexpiBaseModels import DexpiBaseModel, DexpiDataTypeBaseModel


def get_composition_attributes(dexpi_object: DexpiBaseModel) -> dict:
    """
    Retrieve attributes from a DEXPI object with the 'composition' attribute category.

    Parameters
    ----------
    dexpi_object : DexpiBaseModel
        An instance of a DEXPI model from which composition attributes are extracted.

    Returns
    -------
    dict
        A dictionary where the keys are field names and the values are the corresponding
        field values with the '"composition"' category."""
    return _get_attributes_with_category(dexpi_object, "composition")


def get_reference_attributes(dexpi_object: DexpiBaseModel) -> dict:
    """
    Retrieve attributes from a DEXPI object with the 'reference' attribute category.

    Parameters
    ----------
    dexpi_object : DexpiBaseModel
        An instance of a DEXPI model from which reference attributes are extracted.

    Returns
    -------
    dict
        A dictionary where the keys are field names and the values are the corresponding
        field values with the '"reference"' category."""
    return _get_attributes_with_category(dexpi_object, "reference")


def get_data_attributes(dexpi_object: DexpiBaseModel) -> dict:
    """
    Retrieve attributes from a DEXPI object with the 'data' attribute category.

    Parameters
    ----------
    dexpi_object : DexpiBaseModel
        An instance of a DEXPI model from which data attributes are extracted.

    Returns
    -------
    dict
        A dictionary where the keys are field names and the values are the corresponding
        field values with the '"data"' category."""
    return _get_attributes_with_category(dexpi_object, "data")


def _get_attributes_with_category(
    dexpi_object: DexpiBaseModel | DexpiDataTypeBaseModel, category: str
) -> dict:
    """
    Retrieve a dictionary of field names and their values from a dexpi_object
    based on the attribute_category annotation in the 'json_schema_extra'.

    This function iterates over the fields of the given dexpi_object,
    checks the 'json_schema_extra' metadata for a key '"attribute_category"',
    and includes the field in the output dictionary if its value matches the
    provided attribute category.

    Parameters
    ----------
    dexpi_object : DexpiBaseModel | DexpiDataTypeBaseModel
        A pydexpi object from which attributes are extracted.
    category : str
        The value of the '"attribute_category"' to filter the fields by.

    Returns
    -------
    dict
        A dictionary where the keys are field names and the values are the
        corresponding field values from the model, filtered by the given
        annotation.
    """

    attribute_dict = {}
    for fld_name, field in dexpi_object.__class__.model_fields.items():
        if field.json_schema_extra is not None:
            if "attribute_category" in field.json_schema_extra:
                if field.json_schema_extra["attribute_category"] == category:
                    attribute_dict[fld_name] = getattr(dexpi_object, fld_name)
    return attribute_dict


def get_dexpi_class(class_name: str) -> DexpiBaseModel:
    """
    Retrieve a DEXPI class by its name.

    Parameters
    ----------
    class_name : str
        The name of the DEXPI class to retrieve.

    Returns
    -------
    DexpiBaseModel
        The DEXPI class with the given name.
    """
    for submodule_name in dir(dexpi_classes):
        # Get the submodule dynamically
        submodule = getattr(dexpi_classes, submodule_name)
        cls = getattr(submodule, class_name, None)
        if cls:
            break
    else:
        raise AttributeError(f"Class {class_name} not a DEXPI class.")
    return cls


def get_dexpi_class_from_uri(uri: str) -> DexpiBaseModel:
    """
    Retrieve a DEXPI class from its URI.

    Parameters
    ----------
    uri : str
        The URI of the DEXPI class to retrieve.

    Returns
    -------
    DexpiBaseModel
        The DEXPI class corresponding to the given URI.
    """
    class_name = uri.split("/")[-1]

    # Strip .py from the back of the class name if it exists and capitalize first letter
    if class_name.endswith(".py"):
        class_name = class_name[:-3]
    class_name = class_name[0].upper() + class_name[1:]
    return get_dexpi_class(class_name)



"""Generation history to store the generation steps of a generator."""

import json
from dataclasses import dataclass, field

from pydexpi.syndata.generator_step import GeneratorStepDict, WriteableStep


@dataclass
class GenerationHistory:
    """Class to store the generation history of a generator.

    Attributes
    ----------
    history : list[GeneratorStepDict]
        The history of the generator.
    """

    history: list[GeneratorStepDict] = field(default_factory=list)

    def write_step(self, step: WriteableStep) -> None:
        """Write a step to the history.

        Parameters
        ----------
        step : WriteableStep
            The step to write to the history.

        Returns
        -------
        None
        """
        self.history.append(step.to_dict())


def save_generation_history_as_json(generation_history: GenerationHistory, path: str) -> None:
    """Save the history as a json file.

    Parameters
    ----------
    generation_history : GenerationHistory
        The history to save.
    path : str
        The path to save the json file to.
    """

    with open(path, "w") as f:
        json.dump(generation_history.history, f, indent=4)


def load_generation_history_from_json(path: str) -> list[GeneratorStepDict]:
    """Load the history from a json file.

    Parameters
    ----------
    path : str
        The path to load the json file from.

    Returns
    -------
    list[GeneratorStepDict]
        The loaded history.
    """
    with open(path) as f:
        history = json.load(f)
    return history



import pickle
from abc import ABC, abstractmethod
from pathlib import Path

from pydexpi.dexpi_classes.dexpiModel import DexpiModel


class Serializer(ABC):
    """Abstract class to (de)serialize DEXPI models."""

    @abstractmethod
    def save(self, model: DexpiModel, dir_path: Path, filename: str):
        """Abstract method to save a DEXPI model to a file.

        Parameters
        ----------
        model : DexpiModel
            DEXPI model that should be saved.
        dir_path : Path
            Directory where the DEXPI model should be saved.
        filename : str
            Filename for the saved DEXPI model.
        """
        pass

    @abstractmethod
    def load(self, dir_path: Path, filename: str):
        """Abstract method to load a DEXPI PID data model from a P&ID file.

        Parameters
        ----------
        dir_path : Path
            Directory where the P&ID file is stored.
        filename : str
            Filename of the P&ID file.

        Returns
        -------
        DexpiModel
            Loaded DEXPI model.
        """
        pass


class PickleSerializer(Serializer):
    def save(self, model: DexpiModel, dir_path: Path, filename: str):
        """Saves a DEXPI model to pickle.

        Parameters
        ----------
        model : DexpiModel
            DEXPI model that should be saved.
        dir_path : Path
            Directory where the DEXPI model should be saved.
        filename : str
            Filename for the saved DEXPI model.
        """
        if not filename.endswith(".pkl"):
            filename += ".pkl"
        path = Path(dir_path) / filename

        with open(path, "wb") as f:
            pickle.dump(model, f)

    def load(self, dir_path: Path, filename: str):
        """Load a DEXPI model from a pickle file.

        Parameters
        ----------
        dir_path : Path
            Directory where the pickle file is stored.
        filename : str
            Filename of the pickle file.

        Returns
        -------
        DexpiModel
            Loaded DEXPI model.
        """
        if not filename.endswith(".pkl"):
            filename += ".pkl"
        path = Path(dir_path) / filename

        with open(path, "rb") as f:
            model = pickle.load(f)
        return model



from .graph_loader import NXGraphLoader
from .json_serializer import JsonSerializer
from .ml_graph_loader import MLGraphLoader
from .proteus_serializer import ProteusSerializer
from .serializer import PickleSerializer



from .pydantic_classes import (
    ActuatingElectricalFunction,
    ActuatingElectricalLocation,
    ActuatingElectricalSystem,
    ActuatingFunction,
    ActuatingSystem,
    ControlledActuator,
    CustomActuatingElectricalSystemComponent,
    CustomActuatingSystemComponent,
    CustomProcessSignalGeneratingSystemComponent,
    ElectronicFrequencyConverter,
    FlowDetector,
    FlowInSignalOffPageConnector,
    FlowOutSignalOffPageConnector,
    InlinePrimaryElementReference,
    InstrumentationLoopFunction,
    MeasuringLineFunction,
    OfflinePrimaryElement,
    OperatedValveReference,
    Positioner,
    PrimaryElement,
    ProcessControlFunction,
    ProcessInstrumentationFunction,
    ProcessSignalGeneratingFunction,
    ProcessSignalGeneratingSystem,
    SensingLocation,
    SignalConveyingFunction,
    SignalConveyingFunctionSource,
    SignalConveyingFunctionTarget,
    SignalLineFunction,
    SignalOffPageConnector,
    SignalOffPageConnectorObjectReference,
    SignalOffPageConnectorReference,
    SignalOffPageConnectorReferenceByNumber,
    Transmitter,
)



import networkx as nx
import pytest

from pydexpi.loaders import MLGraphLoader


@pytest.fixture
def graph_loader(loaded_example_dexpi):
    dexpi_model = loaded_example_dexpi
    my_graph_loader = MLGraphLoader()
    my_graph_loader.dexpi_to_graph(dexpi_model)
    return my_graph_loader


# @pytest.mark.dependency(depends=["loaded_example_graph"])
def test_graph_stats(graph_loader):
    """Test number of nodes and edges."""
    # Number of edges
    loaded_example_graph = graph_loader.plant_graph
    assert loaded_example_graph.number_of_edges() == 36

    # Number of nodes
    assert loaded_example_graph.number_of_nodes() == 33


# @pytest.mark.dependency(depends=["loaded_example_graph"])
def test_graph_integrity(graph_loader):
    """Test if the graph is completely connected."""
    loaded_example_graph = graph_loader.plant_graph
    # Number of independent subgraphs
    UG = nx.to_undirected(loaded_example_graph)
    sub_graphs = [UG.subgraph(c).copy() for c in nx.connected_components(UG)]
    assert len(sub_graphs) == 1


# @pytest.mark.dependency(depends=["loaded_example_graph"])
def test_graph_format(graph_loader):
    """Test the graph format using internal validation."""
    graph_loader.validate_graph_format()



import datetime
from collections import Counter
from copy import deepcopy

from pydexpi.dexpi_classes import (
    customization,
    equipment,
    metaData,
    piping,
)
from pydexpi.toolkits import model_toolkit as mt


def test_combine_dexpi_models(simple_dexpi_model_factory):
    """Test combine two or more dexpi models"""

    # Retrieve/set some data for later assertion
    metadata = metaData.MetaData()
    export_date_time = datetime.datetime(2020, 10, 3)
    models = [deepcopy(simple_dexpi_model_factory()) for i in range(3)]

    new_model = mt.combine_dexpi_models(models, metaData=metadata, exportDateTime=export_date_time)

    assert len(new_model.conceptualModel.pipingNetworkSystems) == 6
    assert new_model.exportDateTime == export_date_time
    assert new_model.conceptualModel.metaData == metadata


def test_import_model_contents_into_model(simple_dexpi_model_factory):
    """Test importing other models into a dexpi model"""

    # Retrieve/set some data for later assertion
    model = simple_dexpi_model_factory()  # call factory to create a new instance
    metadata = model.conceptualModel.metaData
    export_date_time = datetime.datetime(1967, 4, 27)
    model.exportDateTime = export_date_time

    # Combine models
    models = [deepcopy(simple_dexpi_model_factory()) for i in range(2)]
    mt.import_model_contents_into_model(model, import_models=models)

    assert len(model.conceptualModel.pipingNetworkSystems) == 6
    assert model.exportDateTime == export_date_time
    assert model.conceptualModel.metaData == metadata


def test_get_instances_with_attribute(simple_dexpi_model_factory):
    """Test getting all instances of a pns with a custom attribute and a data
    attribute."""
    # call factory to create a new instance
    model = simple_dexpi_model_factory()

    # First, assert getting an empty list if no attributes are set
    instances = mt.get_instances_with_attribute(model, "segmentNumber", "123")
    assert not instances

    # Assert getting None values
    instances = mt.get_instances_with_attribute(model, "targetItem", None)
    assert len(instances) == 2  # One pipe and one segment

    segment1 = model.conceptualModel.pipingNetworkSystems[0].segments[0]
    equipment1 = model.conceptualModel.taggedPlantItems[0]
    segment1.segmentNumber = "123"

    instances = mt.get_instances_with_attribute(
        model, "segmentNumber", "123", piping.PipingNetworkSegment
    )
    assert instances == [segment1]

    for obj in [segment1, equipment1]:
        new_custom_attribute = customization.CustomAttribute(
            attributeName="Testattribute", value="321"
        )
        obj.customAttributes.append(new_custom_attribute)

    # Rediscover instances
    instances = mt.get_instances_with_attribute(model, "Testattribute", "321")
    assert len(instances) == 2
    assert segment1 in instances
    assert equipment1 in instances

    # Test discovering instances without specifying the target value
    instances = mt.get_instances_with_attribute(model, "Testattribute")
    assert len(instances) == 2
    assert segment1 in instances
    assert equipment1 in instances

    # Rediscover instances, but only the piping network segment
    instances = mt.get_instances_with_attribute(
        model, "Testattribute", "321", piping.PipingNetworkSegment
    )
    assert instances == [segment1]


def test_get_all_instances_in_model(simple_conceptual_model_factory):
    """Test getting all instances of several classes in the simple conceptual
    model"""

    conceptual_model = simple_conceptual_model_factory()  # call factory
    pipes = mt.get_all_instances_in_model(conceptual_model, piping.Pipe)
    assert len(pipes) == 8

    equipments = mt.get_all_instances_in_model(conceptual_model, equipment.Equipment)
    assert len(equipments) == 1

    nozzles = mt.get_all_instances_in_model(equipments[0], equipment.Nozzle)
    assert len(nozzles) == 2

    p_and_e = mt.get_all_instances_in_model(conceptual_model, (equipment.Equipment, piping.Pipe))
    assert len(p_and_e) == len(pipes) + len(equipments)

    everything = mt.get_all_instances_in_model(conceptual_model)
    # Make sure there are no duplicates
    element_counts = Counter(everything)
    duplicates = [item for item, count in element_counts.items() if count > 1]
    assert len(duplicates) == 0
    assert len(everything) == 35


def test_object_is_in_model(simple_dexpi_model_factory):
    """Test to discover an object in a simple dexpi model and cross check with
    an unrelated object."""
    model = simple_dexpi_model_factory()  # call factory to create a new instance
    true_member = model.conceptualModel.pipingNetworkSystems[0].segments[0]
    assert mt.object_is_in_model(model, true_member)
    false_member = piping.Pipe()
    assert not mt.object_is_in_model(model, false_member)



from abc import ABC, abstractmethod

from networkx import Graph

from pydexpi.dexpi_classes.dexpiModel import DexpiModel


class NXGraphLoader(ABC):
    """
    Abstract base class for loading and transforming NX graphs and Dexpi models.

    Provides an interface for parsing between pyDEXPI models and NetworkX graph
    representations.
    """

    @abstractmethod
    def dexpi_to_graph(self, dexpi_model: DexpiModel) -> Graph:
        """
        Parse a pyDEXPI model into a NetworkX graph representation.

        Parameters
        ----------
        pydexpi_model : DexpiModel
            The pyDEXPI model to be parsed into a NetworkX graph.

        Returns
        -------
        nx.Graph
            The parsed NetworkX graph representation of the pyDEXPI model.
        """
        pass

    @abstractmethod
    def graph_to_dexpi(self, plant_graph: Graph) -> DexpiModel:
        """
        Parse a NetworkX graph into a pyDEXPI model representation.

        Parameters
        ----------
        plant_graph : nx.Graph
            The NetworkX graph representation of a plant to be parsed into a Dexpi model.

        Returns
        -------
        DexpiModel
            The parsed pyDEXPI model representation of the plant graph.
        """
        pass

    @abstractmethod
    def validate_graph_format(self, plant_graph: Graph) -> bool:
        """
        Validate that the given NetworkX graph conforms to the expected format.

        This method ensures that the structure, nodes, edges, and associated metadata
        in the graph meet the requirements for parsing or transformation into a
        Dexpi model.

        Parameters
        ----------
        plant_graph : nx.Graph
            The NetworkX graph to validate.

        Returns
        -------
        bool
            True if the graph is valid, False otherwise.
        """
        pass



"""Tests for the module pattern distribution."""

import pytest

from pydexpi.syndata.pattern_distribution import PatternDistribution


def test_distribution_constructor(simple_distribution_factory):
    """Test if a distribution is created correctly"""
    simple_distribution = simple_distribution_factory("simple_distribution")
    assert len(simple_distribution.patterns) == 2


def test_add_pattern_and_normalize(simple_distribution_factory, simple_pattern_factory):
    """Test adding a pattern to the distribution with and without normalization."""
    simple_distribution = simple_distribution_factory("simple_distribution")
    # Add a new pattern without normalization
    the_pattern = simple_pattern_factory("New label")
    simple_distribution.add_pattern(the_pattern, 0.5, normalize=False)
    assert len(simple_distribution.patterns) == 3
    assert sum(simple_distribution.probabilities.values()) == 1.5

    # Add a new pattern with normalization
    simple_distribution.add_pattern(simple_pattern_factory("New label 2"), 0.5)
    assert 0.99999 < sum(simple_distribution.probabilities.values()) < 1.00001

    # try adding a pattern that is already part
    with pytest.raises(ValueError):
        simple_distribution.add_pattern(the_pattern, 0.5)

    # try adding a pattern with invalid connectors
    invalid_pattern = simple_pattern_factory("New label 3")
    del invalid_pattern.connectors[next(iter(invalid_pattern.connectors))]
    with pytest.raises(ValueError):
        simple_distribution.add_pattern(invalid_pattern, 0.5)


def test_check_pattern_compatibility(simple_distribution_factory, simple_pattern_factory):
    """Test the check pattern compatibility method"""
    simple_distribution = simple_distribution_factory("simple_distribution")
    the_pattern = simple_pattern_factory("New label")
    assert simple_distribution.check_pattern_compatibility(the_pattern)
    del the_pattern.connectors[next(iter(the_pattern.connectors))]
    assert not simple_distribution.check_pattern_compatibility(the_pattern)


def test_sample_pattern(simple_distribution_factory):
    """Test if the simple distributions sampling methods work"""
    simple_distribution = simple_distribution_factory("simple_distribution")
    assert simple_distribution.check_pattern_compatibility(simple_distribution.sample_pattern()[0])
    assert simple_distribution.check_pattern_compatibility(simple_distribution.random_pattern()[0])
    for pattern in simple_distribution:
        assert simple_distribution.check_pattern_compatibility(pattern[0])


def test_save_and_load_distribution(tmp_path, simple_distribution_factory):
    """Test loading and saving a distribution"""
    simple_distribution = simple_distribution_factory("simple_distribution")
    simple_distribution.save(tmp_path)
    loaded_distribution = PatternDistribution.load(tmp_path, simple_distribution.name)
    assert simple_distribution.name == loaded_distribution.name
    for key in simple_distribution.labels():
        assert simple_distribution.patterns[key].label == loaded_distribution.patterns[key].label



from collections.abc import Callable

import networkx as nx
import pytest

from pydexpi.dexpi_classes.pydantic_classes import DexpiModel, PipingNetworkSegment
from pydexpi.loaders.json_serializer import JsonSerializer
from pydexpi.loaders.ml_graph_loader import MLGraphLoader


def test_json_loader_dict_simple_pns(simple_pns_factory: Callable[[], PipingNetworkSegment]):
    """Test if the JSONLoader can convert a simple PNS to dict and back."""

    json_loader = JsonSerializer()
    simple_pns = simple_pns_factory()

    # Convert PNS to dict
    pns_dict = json_loader.model_to_dict(simple_pns)

    # Convert dict back to PNS
    reconstructed_pns = json_loader.dict_to_model(pns_dict)

    # Check if the original and reconstructed PNS have equal features
    assert len(simple_pns.items) == len(reconstructed_pns.items)
    assert len(simple_pns.connections) == len(reconstructed_pns.connections)
    assert simple_pns.items[0].nodes[0].id == reconstructed_pns.items[0].nodes[0].id

    # Compare via identity, which references the object ids
    assert simple_pns.connections[1].sourceItem == reconstructed_pns.connections[1].sourceItem
    assert simple_pns.connections[1].targetItem == reconstructed_pns.connections[1].targetItem
    assert simple_pns.connections[1].sourceNode == reconstructed_pns.connections[1].sourceNode


def test_json_loader_on_full_model(loaded_example_dexpi: DexpiModel):
    """Test if the JSONLoader can convert a full Dexpi model to dict and back."""

    json_loader = JsonSerializer()

    # Convert Dexpi model to dict
    json_dict = json_loader.model_to_dict(loaded_example_dexpi)

    # Convert dict back to Dexpi model
    reconstructed_model = json_loader.dict_to_model(json_dict)

    # Compare via graph export
    gr_loader = MLGraphLoader()
    orig_graph = gr_loader.dexpi_to_graph(loaded_example_dexpi)
    recon_graph = gr_loader.dexpi_to_graph(reconstructed_model)

    # Check if the original and reconstructed graphs are isomorphic
    is_isomorphic = nx.is_isomorphic(
        orig_graph,
        recon_graph,
        node_match=lambda n1, n2: n1 == n2,
        edge_match=lambda e1, e2: e1 == e2,
    )
    assert is_isomorphic, "The original and reconstructed graphs are not isomorphic."

    # Check via the json export. It should be the same as the original dict.
    reconstructed_json_dict = json_loader.model_to_dict(reconstructed_model)
    assert json_dict == reconstructed_json_dict, (
        "The original and reconstructed JSON dicts do not match."
    )


def test_load_save_json(loaded_example_dexpi: DexpiModel, tmp_path: str):
    """Test if the JSONLoader can save and load a Dexpi model correctly."""

    json_loader = JsonSerializer()

    # Save the dict to a file
    json_loader.save(loaded_example_dexpi, tmp_path, "test_model.json")

    # Load the dict from the file
    reconstructed_model = json_loader.load(tmp_path, "test_model.json")

    # Compare via graph export
    gr_loader = MLGraphLoader()
    orig_graph = gr_loader.dexpi_to_graph(loaded_example_dexpi)
    recon_graph = gr_loader.dexpi_to_graph(reconstructed_model)

    # Check if the original and reconstructed graphs are isomorphic
    is_isomorphic = nx.is_isomorphic(
        orig_graph,
        recon_graph,
        node_match=lambda n1, n2: n1 == n2,
        edge_match=lambda e1, e2: e1 == e2,
    )
    assert is_isomorphic, "The original and reconstructed graphs are not isomorphic."

    # Test for non-existent file ending
    json_loader.save(loaded_example_dexpi, tmp_path, "test_model2")

    # Load the dict from the file and check that no error is raised
    _ = json_loader.load(tmp_path, "test_model2.json")
    _ = json_loader.load(tmp_path, "test_model2")

    # Test loadin a non-existent file
    with pytest.raises(FileNotFoundError):
        json_loader.load(tmp_path, "non_existent_file.json")



"""This module contains the abstract classes of pattern and pattern connectors
for synthetic data generation. To use these, extend the abstract classes to
wrap the data classes of interest and define connection and incorporation logic.
"""

from __future__ import annotations

import copy
import pickle
from abc import ABC, abstractmethod
from pathlib import Path


class Connector(ABC):
    """
    Abstract base class for a connector that manages connection interfaces for
    synthetic data patterns, including connection logic. The class also manages
    a set of observer connectors for parallel generation.

    To implement this abstract class, define the data object to be wrapped by
    overriding the constructor and implement the connection logic in the
    abstract _implement_connection method.

    Attributes
    ----------
    label : str
        A label representing the connector.
    kwinfos : dict, optional
        Additional keyword information that can be accessed, manipulated, and
        referenced by the generator during generation.
    _is_active : bool
        Internal security flag indicating whether the connector is active or
        has been connected or deactivated.
    """

    def __init__(self, label: str, kwinfos: dict = None) -> None:
        """
        Initializes the connector object with a label and additional information
        as kwargs.

        Parameters
        ----------
        label : str
            A label for the connector instance.
        **kwinfos
            Additional keyword arguments representing metadata for the
            connector.
        """
        self.label = label
        self.kwinfos = kwinfos if kwinfos is not None else {}
        # Include a safety switch to make sure a connector is only connected once
        self._is_active = True

    @property
    def is_active(self) -> bool:
        """Returns the active status of the connector."""
        return self._is_active

    def connect_to_counterpart(self, counterpart: Connector) -> None:
        """
        Establishes the connection between this connector and a counterpart
        connector.

        This method first checks whether this connector is still active and
        raises a RuntimeError if not. Then, the method calls the subclass-
        specific logic defined in `_implement_connection()` to establish the
        primary connection.

        Parameters
        ----------
        counterpart : Connector
            The counterpart connector instance to which this connector
            will be connected.

        Raises
        ------
        RuntimeError
            If this connector is no longer active.
        ValueError
            If the counterpart is the same object as self or the counterpart is
            invalid.
        """
        # If this connector has been connected before, raise an Error
        if not self._is_active:
            msg = "Attempted to connect a connector object that was already deactivated."
            raise RuntimeError(msg)

        # Make sure connector is not connected to itself
        if counterpart == self:
            raise ValueError("Cannot connect a counterpart to itself.")

        # Make sure connector is valid
        if not self.assess_valid_counterpart(counterpart):
            raise ValueError("Counterpart connector is not valid for connection.")

        # Call subclass specific connection logic
        self._implement_connection(counterpart)

        # Set is_active tag to false after successful connection
        self._is_active = False
        counterpart._is_active = False

    def deactivate(self) -> None:
        """Deactivates an unconnected connector.

        Sets the active flag to False and calls optional subclass specific deactivation logic.

        Raises
        ------
        RuntimeError
            If this connector is no longer active.
        """
        # If this connector has been connected before, raise an Error
        if not self._is_active:
            msg = "Attempted to connect a connector object that was already deactivated."
            raise RuntimeError(msg)

        # Call subclass specific deactivation logic
        self._implement_deactivation()

        # Set is_active tag to false after successful deactivation
        self._is_active = False

    @abstractmethod
    def assess_valid_counterpart(self, counterpart: Connector) -> bool:
        """Abstract method to assess if a given Counterpart is valid for
        connection. Subclass-specific logic to be implemented by child classes.

        Parameters
        ----------
        counterpart : Connector
            The candidate counterpart to be assessed for connection
            compatibility.

        Returns
        -------
        bool
            True if counterpart is suitable for connection, False otherwise
        """
        pass

    @abstractmethod
    def _implement_connection(self, counterpart: Connector) -> None:
        """
        Abstract method to implement the primary connection logic for a
        subclass.

        This method is intended to be overridden by subclasses to define how
        the main connection between this connector and the counterpart connector
        should be established, depending on the wrapped interface data
        structure. It is called after validation but before the observers
        are connected.

        Parameters
        ----------
        counterpart : Connector
            The counterpart connector instance to be connected to.
        """
        pass

    def _implement_deactivation(self) -> None:
        """Optional abstract method executed when the connector is deactivated without connection."""
        pass


class Pattern(ABC):
    """
    Abstract base class for a pattern that wraps synthetic data patterns, and
    contains pattern interfaces via a list of connections. The class also
    manages a set of observer patterns for parallel generation.

    To implement this abstract class, define the data object to be wrapped by
    overriding the constructor and implement the incorporation logic in the
    abstract `_implement_incorporation method`. If, in an implementation of
    the Pattern, connections need to be performed on a Pattern level and not on
    a Connector level, the _connect_via_pattern method can be implemented.

    Attributes
    ----------
    label : str
        A label representing the pattern.
    connectors : dict[str, Connector]
        The connectors of the pattern defining the connection interfaces
    observer_patterns : dict, optional
        A dictionary of observer patterns, where keys are observer tags and
        values are pattern instances.
    kwinfos : dict, optional
        Additional keyword information that can be accessed, manipulated, and
        referenced by the generator during generation.
    _is_incorporated : bool
        Internal security flag indicating whether the pattern has been
        incorporated.
    """

    def __init__(
        self,
        label: str,
        connectors: dict[str, Connector],
        observer_patterns: dict[str, Pattern] = None,
        kwinfos: dict = None,
    ) -> None:
        self.label = label

        self._connectors = None
        self.connectors = connectors

        self.observer_patterns: dict[str, Pattern] = {}
        self._is_incorporated = False
        self.kwinfos = kwinfos if kwinfos is not None else {}
        if observer_patterns is not None:
            for tag, observer in observer_patterns.items():
                self.add_observer(tag, observer)

    @property
    def is_incorporated(self) -> bool:
        """Returns the incorporation status of the pattern."""
        return self._is_incorporated

    @property
    def connectors(self) -> dict[str, Connector]:
        """Returns the connectors of the pattern."""
        return self._connectors

    @connectors.setter
    def connectors(self, new_connectors: dict[str, Connector] | list[Connector]) -> None:
        """Sets the connectors of the pattern.

        This method can be used to update the connectors of the pattern. The new connectors can be
        provided as a dictionary where the keys are the labels and the values are the Connector
        objects. Alternatively, a list of Connector objects can be provided.


        Parameters
        ----------
        new_connectors : dict[str, Connector]|list[Connector]
            The new connectors to be set. Can be a list of Connector objects or
            a dictionary where keys are the labels and values are the Connector
            objects.

        Raises
        ------
        ValueError
            If the keys of the dictionary do not match the labels of the
            connectors.
        """
        if isinstance(new_connectors, list):
            self._connectors = {connector.label: connector for connector in new_connectors}
        elif isinstance(new_connectors, dict):
            if not all(key == value.label for key, value in new_connectors.items()):
                raise ValueError("Keys of the dictionary must match the labels of the connectors.")
            self._connectors = new_connectors

    def add_observer(self, observer_tag: str, new_observer: Pattern) -> None:
        """
        Adds a new observer pattern to this pattern instance. To ensure
        conceptual correctness in generation logic, observers must have the same
        string for their label attribute, and the labels of the observer's
        connectors must correspond to the labels of the subject connectors

        Parameters
        ----------
        observer_tag : str
            The identifier for the observer pattern.
        new_observer : Connector
            The new observer pattern to be added.

        Raises
        ------
        ValueError
            If the new observer's label does not match this subject's label, or
            if the connector's of the patterns do not correspond via their
            labels.
        RuntimeError
            If this pattern is already incorporated in another pattern.
        """
        if self._is_incorporated:
            raise RuntimeError("Pattern is already incorporated")
        # Check if observer label matches subect label
        if new_observer.label != self.label:
            msg = "Invalid observer pattern: Observer has a different label."
            raise ValueError(msg)
        # Check if observer has the same number of connectors
        if len(self.connectors) != len(new_observer.connectors):
            msg = "New observer does not have the same amount of connectors"
            raise ValueError(msg)
        # Check if observer connectors match the connectors
        for label, own_connector in self.connectors.items():
            observer_connector = new_observer.connectors.get(label)
            if observer_connector is None or own_connector.label != observer_connector.label:
                msg = "Connectors of the new observer do not match the subject connectors."
                raise ValueError(msg)

        # Add observer, and add each connector of observer as observer to
        # subject connectors
        self.observer_patterns[observer_tag] = new_observer

    def incorporate_pattern(
        self,
        own_connector: Connector,
        counterpart: Pattern,
        counterpart_connector: Connector,
    ) -> None:
        """
        Incorporates a counterpart pattern, including by connecting specified
        connectors and importing of data.

        This method first checks whether the connector arguments are part of
        the corresponding patterns. Then, the method calls the subclass-specific
        logic defined in `_implement_incorporation()` to import the data. Then
        connectors are connected via connect_to_counterpart in the Connectors or
        the alternative connection on pattern level defined in
        _connect_via_pattern. Once the connections are completed, the
        connectors involved in the connection (`own_connector` and
        `counterpart_connector`) are removed from the pattern. After
        incorporation, the observers are notified to do the same.

        Parameters
        ----------
        own_connector : Connector
            The connector instance associated with this pattern that will be
            responsible for the connection.
        counterpart : Pattern
            The counterpart pattern instance which will be incorporated.
        counterpart_connector : Connector
            The connector instance associated with the counterpart pattern that
            will be used for the connection.

        Raises
        ------
        ValueError
            If the connectors for the connection are not associated to the
            respective patterns, or if the observers of the counterpart doesn't
            match with this object's observers.
        ValueError
            From connector arguments: If the types of the connectors are an
            invalid combination.
        RuntimeError
            If this pattern is already incorporated in another pattern.
        """
        if self._is_incorporated:
            raise RuntimeError("Pattern is already incorporated")

        # Make sure that all counterpart connector keys are not already in the pattern connectors,
        # except for the own_connector and the counterpart_connector. that are removed in this
        # operation.
        for key, connector in counterpart.connectors.items():
            if connector == counterpart_connector:
                continue
            if key == own_connector.label:
                continue
            if key in self.connectors.keys():
                raise ValueError(f"Key {key} already in this pattern's connectors.")

        # Make sure that connectors are correctly associated and valid
        if own_connector not in self.connectors.values():
            raise ValueError("Connector argument own_connector not in this pattern's connectors.")
        if counterpart_connector not in counterpart.connectors.values():
            raise ValueError(
                "Connector argument counterpart_connector not in the counterpart pattern's connectors."
            )
        if set(self.observer_patterns.keys()) != set(counterpart.observer_patterns.keys()):
            msg = "Pattern doesnt share the same observers with counterpart pattern."
            raise ValueError(msg)
        if not own_connector.assess_valid_counterpart(counterpart_connector):
            raise ValueError("Own and counterpart connector are not valid for connection.")

        # Perform subtype specific incorporation
        self._implement_incorporation(counterpart)

        # Connect the connectors via connectors, or alternatively via the
        # optional connection logic defined on a pattern level
        own_connector.connect_to_counterpart(counterpart_connector)
        self._connect_via_pattern(own_connector, counterpart_connector)

        # Manage connectors. Add all connectors of the counterpart that are not the
        # counterpart_connector. Then, remove the own_connector from the pattern.
        for key, connector in counterpart.connectors.items():
            if connector != counterpart_connector:
                self.connectors[key] = connector

        del self.connectors[own_connector.label]

        # Inform observers
        for tag, observer in self.observer_patterns.items():
            counterpart_observer = counterpart.observer_patterns[tag]
            own_connector_observer = observer.connectors[own_connector.label]
            counterpart_connector_observer = counterpart_observer.connectors[
                counterpart_connector.label
            ]
            observer.incorporate_pattern(
                own_connector_observer,
                counterpart_observer,
                counterpart_connector_observer,
            )
        counterpart._is_incorporated = True

    def connect_internal(self, connector: Connector, counterpart: Connector) -> None:
        """Connects two connectors of the pattern internally, e.g., for recycle
        or auxiliary connections.

        First, ensures that both connector arguments are part of this pattern.
        Then, connects the connectors and removes them from this patterns open
        connectors. Finally, observers are informed and their connect_internal
        method is invoked.

        Parameters
        ----------
        connector : Connector
            The first connector to be connected.
        counterpart : Connector
            The second connector to be connected to the first.

        Raises
        ------
        RuntimeError
            If the pattern is already incorporated.
        ValueError
            If either of the connectors provided is not in this pattern's
            connectors.
        """
        # First make sure that this pattern has not yet been incorporated
        if self._is_incorporated:
            raise RuntimeError("Pattern is already incorporated.")

        # Make sure that connectors are correctly associated
        if (connector not in self.connectors.values()) or (
            counterpart not in self.connectors.values()
        ):
            raise ValueError("A connector argument is not in this patterns connectors.")

        # Perform connection via connectors and/or via pattern
        connector.connect_to_counterpart(counterpart)
        self._connect_via_pattern(connector, counterpart)

        # Drop the newly connected connectors
        del self.connectors[connector.label]
        del self.connectors[counterpart.label]

        # Inform observers
        for observer in self.observer_patterns.values():
            observer_connector = observer.connectors[connector.label]
            observer_counterpart = observer.connectors[counterpart.label]
            observer.connect_internal(observer_connector, observer_counterpart)

    def drop_connector(self, connector: Connector) -> None:
        """Removes and deactivates an active, unconnected connector from the pattern.

        First, the method checks if the pattern is not yet incorporated and if
        the connector part of this pattern. Then, it calls the deactivate method
        of the connector and removes it from this pattern's connectors. Finally,
        observers are informed to also drop the corresponding connector.

        Used e.g if connector unconnected at the end of the generation process

        Parameters
        ----------
        connector : Connector
            The connector to be dropped.

        Raises
        ------
        ValueError
            If the connector is not in this pattern's connectors.
        RuntimeError
            If this pattern is already incorporated.
        """
        # Make sure pattern is not incorporated and connector is part of
        # connectors
        if self._is_incorporated:
            raise RuntimeError("Pattern is already incorporated")
        if connector not in self.connectors.values():
            raise ValueError("Connector argument not in this patterns connectors")

        # Drop connector
        connector.deactivate()
        del self.connectors[connector.label]

        # Inform observers
        for observer in self.observer_patterns.values():
            observer_connector = observer.connectors[connector.label]
            observer.drop_connector(observer_connector)

    def relabel_connector(self, connector: Connector, new_label: str) -> None:
        """Relabels a connector and informs all observers to do the same.

        Parameters
        ----------
        connector : Connector
            The connector to be relabeled.
        new_label : str
            The new label for the connector.
        """
        # Change label of the connector
        old_label = connector.label
        connector.label = new_label
        self.connectors[new_label] = connector
        del self.connectors[old_label]

        # Inform observers
        for observer in self.observer_patterns.values():
            observer_connector = observer.connectors[old_label]
            observer.relabel_connector(observer_connector, new_label)

    def copy_pattern(self) -> Pattern:
        """Make a copy of the pattern.

        Copy operation to be applied when sampling and incorporating a
        Pattern, so that the template pattern can be sampled and used
        multple times during generation.

        By default, patterns are deepcopied with copy.deepcopy, and
        `copy_pattern()` is invoked on the observers. If more or more complex
        operations are required for this, such as reassigning ids, the method
        should be overridden appropriately.

        Returns
        -------
        Pattern
            The copied pattern"""
        # Temporarily detach observers
        temp_observer_patterns = self.observer_patterns
        self.observer_patterns = {}

        # Copy pattern, but make sure to reattach observers in case of a problem
        try:
            the_copied_pattern = copy.deepcopy(self)
        finally:
            self.observer_patterns = temp_observer_patterns

        # Finally, invoke copy_pattern on all observers and add them as
        # observers to the copy

        for observer_tag, observer in temp_observer_patterns.items():
            copied_observer = observer.copy_pattern()
            the_copied_pattern.add_observer(observer_tag, copied_observer)

        return the_copied_pattern

    def change_label(self, new_label: str) -> None:
        """This method updates the label of the object and notifies
        all observer patterns associated to also change the label.

        Parameters
        ----------
        new_label : str
            The new label to assign to the object.
        """
        # Change label
        self.label = new_label

        # Inform observers
        for observer in self.observer_patterns.values():
            observer.change_label(new_label)

        self.label = new_label

    def save(self, dir_path: Path, file_name: str) -> None:
        """
        Default saving behavior of an abstract pattern. Saves the current
        pattern to a pickle file.

        This method can be overridden by subclasses to provide more meaningful
        serializations based on the data structure enclosed by the subclass.

        Parameters
        ----------
        dir_path : Path
            The directory path where the pickle file will be saved.
        file_name : str
            The name of the pickle file. If the extension ".pkl" is not included,
            it will be appended automatically.
        """
        if not file_name.endswith(".pkl"):
            file_name += ".pkl"
        path = Path(dir_path) / file_name
        with open(path, "wb") as file:
            # Use pickle to serialize the object
            pickle.dump(self, file)

    @classmethod
    def load(cls, dir_path: Path, file_name: str) -> None:
        """
        Default loading behavior of an abstract pattern. Loads a pattern from a
        pickle file.

        This method can be overridden by subclasses to provide more meaningful
        deserializations based on the data structure enclosed by the subclass.

        Parameters
        ----------
        dir_path : Path
            The directory path where the pickle file is located.
        file_name : str
            The name of the pickle file. If the extension ".pkl" is not included,
            it will be appended automatically.

        Returns
        -------
        cls:
            The loaded pattern, if it is an instance of the expected class.

        Raises
        ------
        OSError:
            If the file does not contain a valid object of the expected class.
        """
        if not file_name.endswith(".pkl"):
            file_name += ".pkl"
        path = Path(dir_path) / file_name
        # Open the file in read-binary mode
        with open(path, "rb") as file:
            # Load the distribution object from the file
            the_pattern = pickle.load(file)

        if not isinstance(the_pattern, cls):
            msg = "Pickle file in path does not contain a valid pattern"
            raise OSError(msg)

        return the_pattern

    @abstractmethod
    def _implement_incorporation(
        self,
        counterpart: Pattern,
    ) -> None:
        """
        Abstract method to implement the primary incorporation logic for a
        subclass.

        This method is intended to be overridden by subclasses to handle the
        incorporation of the data of the counterpart into this pattern,
        depending on the wrapped data structure. It is called after validation
        but before the observers are connected.

        NOTE: The data from the counterpart is meant to be imported into the
        data of this pattern, not copied into a new pattern (therefore, no
        return). This ensures the continued validity of connector references.

        Parameters
        ----------
        counterpart : Pattern
            The counterpart pattern instance which will be incorporated.
        """
        pass

    def _connect_via_pattern(
        self,
        connector: Connector,
        counterpart_connector: Connector,
    ) -> None:
        """This method offers an alternative mode to connect the connectors.
        It can (and should only) be used and implemented if the connection
        requires the pattern-level data structure.

        Parameters
        ----------
        connector : Connector
            The connector instance that should be connected to
            counterpart_connector.
        counterpart_connector : Connector
            The connector instance that should be connected to connector.
        """
        pass



from .pydantic_classes import (
    ChamberFunctionClassification,
    CompositionBreakClassification,
    ConfidentialityClassification,
    DetonationProofArtefactClassification,
    ExplosionProofArtefactClassification,
    FailActionClassification,
    FireResistantArtefactClassification,
    GmpRelevanceClassification,
    GuaranteedSupplyFunctionClassification,
    HeatTracingTypeClassification,
    InsulationBreakClassification,
    JacketedPipeClassification,
    LocationClassification,
    NominalDiameterBreakClassification,
    NominalDiameterStandardClassification,
    NominalPressureStandardClassification,
    NumberOfPortsClassification,
    OnHoldClassification,
    OperationClassification,
    PipingClassArtefactClassification,
    PipingClassBreakClassification,
    PipingNetworkSegmentFlowClassification,
    PipingNetworkSegmentSlopeClassification,
    PortStatusClassification,
    PrimarySecondaryPipingNetworkSegmentClassification,
    QualityRelevanceClassification,
    SignalConveyingTypeClassification,
    SiphonClassification,
)
